import types

import line_profiler

import mepylome  # Import the target module


def profile_all_functions(module):
    """Adds the @profile decorator to all functions in the given module."""
    profiler = line_profiler.LineProfiler()
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        # Profile only functions and bound methods
        if isinstance(
            attr, (types.FunctionType, types.MethodType)
        ) and attr.__module__.startswith("mepylome"):
            profiler.add_function(attr)
    return profiler


def profile_methyldata():
    from pathlib import Path

    from mepylome import MethylData, idat_basepaths

    idat_files = idat_basepaths(
        Path("/data/sarcoma_idat/csa_project/idat_CSA")
    )

    for file_ in idat_files:
        _ = MethylData(file=file_, prep="illumina")


def profile_cnv():
    from mepylome.analysis import MethylAnalysis

    analysis = MethylAnalysis(
        analysis_dir="~/mepylome/tutorial/tutorial_analysis/",
        reference_dir="~/mepylome/tutorial/tutorial_reference/",
        n_cpgs=25000,
        load_full_betas=True,
        overlap=False,
        debug=True,
        do_seg=True,
    )
    # analysis.make_umap()
    analysis.precompute_cnvs()


def main():
    profile_cnv()


if __name__ == "__main__":
    # Add profiling for all functions in the mepylome module
    profiler = profile_all_functions(mepylome)
    profiler.run("main()")
    profiler.print_stats()



"""Pipeline that runs diagnostics on new cases from config and saves reports.

This script executes a diagnostic pipeline defined in a JSON/YAML config file.
It processes IDAT files, applies supervised and unsupervised (UMAP)
classifiers, and saves the resulting reports in the directories containing the
new cases. The pipeline is highly customizable through the config file.
"""

import argparse
import io
import json
import logging
import zipfile
from pathlib import Path

import pandas as pd
import requests
import tomllib

from mepylome import ArrayType, Manifest
from mepylome.analysis import MethylAnalysis
from mepylome.analysis.methyl_clf import make_classifier_report_page
from mepylome.cli import get_app_version
from mepylome.dtypes.manifests import (
    DOWNLOAD_DIR,
    MANIFEST_URL,
    REMOTE_FILENAME,
)

IMG_HEIGHT = 2000
FONTSIZE = 23
IMG_WIDTH = 1000


def load_config(config_path):
    """Load a JSON or YAML file and return its contents as a dictionary."""
    with config_path.open(
        "rb" if config_path.suffix == ".toml" else "r"
    ) as file:
        if config_path.suffix == ".json":
            return json.load(file)
        if config_path.suffix in [".yaml", ".yml"]:
            import yaml

            return yaml.safe_load(file)
        if config_path.suffix == ".toml":
            return tomllib.load(file)
    raise ValueError("Unsupported config format")


def sex_chromosome_cpgs():
    """Returns a set of CpGs on sex chromosomes for all arrays."""
    array_types = ["450k", "epic", "epicv2"]
    sex_cpgs = set()
    for array_type in array_types:
        manifest = Manifest(array_type)
        sex_cpgs.update(
            manifest.data_frame.loc[
                manifest.data_frame.Chromosome.isin([23, 24]), "IlmnID"
            ]
        )
    return sex_cpgs


def generate_default_blacklist(blacklist_path):
    """Returns and caches CpG sites that should be blacklisted."""
    blacklist_path = Path(blacklist_path).expanduser()
    if not blacklist_path.exists():
        print("Generating blacklist. Can take some time...")
        manifest_url = MANIFEST_URL[ArrayType.ILLUMINA_EPIC]
        DOWNLOAD_DIR.mkdir(parents=True, exist_ok=True)
        response = requests.get(manifest_url)
        html_sucess_ok_code = 200
        if response.status_code == html_sucess_ok_code:
            with zipfile.ZipFile(io.BytesIO(response.content)) as thezip:
                thezip.extractall(DOWNLOAD_DIR)
        else:
            msg = f"Failed to download the file: {response.status_code}"
            raise RuntimeError(msg)
        csv_path = DOWNLOAD_DIR / REMOTE_FILENAME[ArrayType.ILLUMINA_EPIC]
        manifest_df = pd.read_csv(csv_path, skiprows=7, low_memory=False)
        flagged_cpgs = manifest_df[
            manifest_df["MFG_Change_Flagged"].fillna(False)
        ]["IlmnID"]
        flagged_cpgs = list(set(flagged_cpgs) | sex_chromosome_cpgs())
        pd.DataFrame(flagged_cpgs).to_csv(
            blacklist_path, index=False, header=False
        )
        csv_path.unlink()
    blacklist_df = pd.read_csv(blacklist_path, header=None)
    return set(blacklist_df.iloc[:, 0])


def unsupervised_classifier(analysis, dataset_name, dataset_config):
    """Generate and save UMAP plots for all 'test_ids'."""

    def generate_and_save_umap(id_):
        """Generate and save UMAP plots for a given sample ID."""
        analysis.cnv_id = id_
        analysis.test_ids = [id_]
        basepath = analysis.idat_handler.test_id_to_path[id_].parent
        base_filename = f"{id_}_{dataset_name}"

        jpg_path = basepath / (base_filename + "_umap.jpg")
        html_path = basepath / (base_filename + "_umap.html")

        if jpg_path.exists() and html_path.exists():
            return

        # Save UMAP to disk
        analysis.make_umap()
        analysis.umap_plot.write_image(
            jpg_path,
            format="jpg",
            width=IMG_HEIGHT,
            height=IMG_WIDTH,
            scale=1,
        )
        analysis.umap_plot.write_html(html_path)

    ids = analysis.idat_handler.test_ids
    for id_ in ids:
        generate_and_save_umap(id_)


def supervised_classifier(analysis, dataset_name, dataset_config):
    """Run supervised classifiers on test samples and save as HTML reports."""
    ids = analysis.idat_handler.test_ids
    if not ids:
        return
    clf_list = dataset_config["classifier_list"]
    clf_results = analysis.classify(
        ids=ids,
        clf_list=clf_list,
    )
    id_to_report = {id_: [] for id_ in ids}

    for clf_result in clf_results:
        for id_, report in zip(ids, clf_result.reports["html"]):
            id_to_report[id_].append(report)

    app_version = get_app_version()
    clf_version = dataset_config.get("version")
    version_str = f"v{clf_version} " if clf_version else ""
    version_str += f"(Mepylome v{app_version})"
    clf_name = dataset_config.get("name", dataset_name)

    for id_ in ids:
        filename = f"{id_}_{dataset_name}_classifiers.html"
        path = analysis.idat_handler.test_id_to_path[id_].parent / filename
        if path.exists():
            continue
        html_str = make_classifier_report_page(
            reports=id_to_report[id_],
            title=f"{clf_name} Classifier {version_str}",
        )
        path.write_text(html_str)


def make_cnv(analysis, dataset_name, dataset_config):
    """Generates CNV Plot and saves it to corresponding directory."""
    ids = analysis.idat_handler.test_ids
    analysis.precompute_cnvs(ids)
    for id_ in ids:
        save_dir = analysis.idat_handler.test_id_to_path[id_].parent
        jpg_path = save_dir / f"{id_}_{dataset_name}_cnv.jpg"
        html_path = save_dir / f"{id_}_{dataset_name}_cnv.html"
        if jpg_path.exists() and html_path.exists():
            continue
        analysis.make_cnv_plot(id_)
        cnv_plot = analysis.cnv_plot
        cnv_plot.update_layout(
            yaxis_range=[-1.1, 1.1],
            font={"size": FONTSIZE},
            margin={"t": 50},
        )
        cnv_plot.write_image(
            jpg_path,
            format="jpg",
            width=IMG_HEIGHT,
            height=IMG_WIDTH,
            scale=1,
        )
        cnv_plot.write_html(html_path)


def run_single_diagnostics(dataset_name, dataset_config, default_blacklist):
    """Run methylation analysis for a single dataset."""
    init_params = dataset_config.get("methyl_analysis", {})
    init_params.setdefault("cpg_blacklist", default_blacklist)

    analysis = MethylAnalysis(**init_params)

    annotation_column = dataset_config.get("annotation_diagnosis_column")
    if annotation_column:
        analysis.idat_handler.selected_columns = annotation_column

    if dataset_config.get("do_classify"):
        supervised_classifier(analysis, dataset_name, dataset_config)

    if dataset_config.get("do_cnv"):
        make_cnv(analysis, dataset_name, dataset_config)

    if dataset_config.get("do_umap"):
        unsupervised_classifier(analysis, dataset_name, dataset_config)


def run_diagnostics(config_path):
    """Run diagnostics for each dataset based on the provided config."""
    config_dict = load_config(config_path)

    # Choose default CpGs[methyl]  that should be blacklisted
    blacklist_path = config_dict.get("default_blacklist")
    default_blacklist = (
        generate_default_blacklist(blacklist_path) if blacklist_path else None
    )

    for dataset_name, dataset_config in config_dict[
        "diagnostic_config"
    ].items():
        run_single_diagnostics(dataset_name, dataset_config, default_blacklist)


def main():
    """Start diagnostics."""
    parser = argparse.ArgumentParser(
        description="Run methylation diagnostics."
    )
    parser.add_argument(
        "--config",
        type=Path,
        default=Path(__file__).parent / "config.toml",
        help=(
            "Path to the config file (default: 'config.toml' in the current "
            "directory)",
        ),
    )
    args = parser.parse_args()

    try:
        config_path = args.config
        logging.info("Starting diagnostics with config: %s", config_path)
        run_diagnostics(config_path)
        logging.info("Diagnostics completed successfully.")
    except Exception as e:
        logging.error("An error occurred: %s", e)
        raise


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



"""Basic usage of the package."""

from pathlib import Path

from mepylome import CNV, MethylData

# Sample
analysis_dir = Path("/path/to/idat/directory")
sample_file = analysis_dir / "200925700125_R07C01"

# CNV neutral reference files
reference_dir = Path("/path/to/reference/directory")

# Get methylation data
sample_methyl = MethylData(file=sample_file)
reference_methyl = MethylData(file=reference_dir)

# Beta value
betas = sample_methyl.betas

# Print overview of processed data
print(sample_methyl)

# CNV anylsis
cnv = CNV.set_all(sample_methyl, reference_methyl)

# Visualize CNV in the browser
cnv.plot()



"""Utilities for handling file operations.

This module provides utilities such as downloading files (used to download
manifest files from Illumina), ensuring directories exist, and working with
file-like objects and archives.
"""

import gzip
import logging
import shutil
import zipfile
from itertools import repeat
from multiprocessing import Pool
from pathlib import Path, PurePath

import pandas as pd
from tqdm import tqdm

logger = logging.getLogger(__name__)

try:
    from importlib.resources import files
except (ImportError, ModuleNotFoundError):
    from importlib_resources import files


__all__ = [
    "download_file",
    "ensure_directory_exists",
    "get_file_object",
    "get_resource_path",
    "get_csv_file",
    "reset_file",
]

GEO_BASE_URL = (
    "https://www.ncbi.nlm.nih.gov/geo/download/"
    "?acc={acc}&format=file&file={filename}"
)


def get_resource_path(package, resource_name=""):
    """Returns the full path to the resource within the specified package."""
    package_path = files(package)
    return package_path.joinpath(resource_name)


def ensure_directory_exists(path_like):
    """Ensures the ancestor directories of the provided path exist."""
    Path(path_like).mkdir(parents=True, exist_ok=True)


def download_file(url, save_path, overwrite=False, show_progress=True):
    """Download a file from a URL and save it to a destination directory.

    Args:
        url (str): The URL from which the file will be downloaded.
        save_path (path_like): The path where the file will be saved.
        overwrite (bool, optional): If True, overwrite the file if it already
            exists. Defaults to False.
        show_progress (bool, optional): If True, displays logging messages and
            progress bar during download. Defaults to True.
    """
    save_path = Path(save_path)
    ensure_directory_exists(save_path.parent)

    if save_path.exists() and not overwrite:
        if show_progress:
            logger.info(
                "File already exists at %s. Skipping download.", save_path
            )
        return

    if show_progress:
        logger.info("Downloading from %s to %s...", url, save_path)

    try:
        import requests

        response = requests.get(url, stream=True)
        response.raise_for_status()

        total_size = int(response.headers.get("content-length", 0))
        block_size = 8192
        progress_bar = (
            tqdm(
                total=total_size,
                unit="iB",
                unit_scale=True,
                desc="Downloading",
            )
            if show_progress and total_size > 0
            else None
        )

        with save_path.open("wb") as file:
            for chunk in response.iter_content(chunk_size=block_size):
                if chunk:
                    file.write(chunk)
                    if progress_bar:
                        progress_bar.update(len(chunk))
        if progress_bar:
            progress_bar.close()

        if show_progress:
            logger.info("Download completed: %s", save_path)

    except requests.RequestException as error:
        msg = f"Failed to download file from {url}. Error: {error}"
        raise requests.RequestException(msg) from error


def download_geo_idat(probe_id, color_channel, output_dir):
    """Downloads an IDAT file for a specific probe and color channel."""
    geo_id, sentrix_id = probe_id.split("_", 1)
    file_name = f"{geo_id}_{sentrix_id}_{color_channel}.idat.gz"
    encoded_file_name = file_name.replace("_", "%5F").replace(".", "%2E")
    url = GEO_BASE_URL.format(acc=geo_id, filename=encoded_file_name)
    save_path = output_dir / file_name
    download_file(url, save_path, show_progress=False)


def download_geo_idat_pair(probe_info):
    """Downloads both 'Grn' and 'Red' IDAT files for a single probe."""
    probe_id, output_dir = probe_info
    for color_channel in ["Grn", "Red"]:
        download_geo_idat(probe_id, color_channel, output_dir)


def download_geo_samples(output_dir, probe_ids):
    """Downloads IDAT file pairs from GEO."""
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    with Pool() as pool:
        list(
            tqdm(
                pool.imap(
                    download_geo_idat_pair,
                    zip(probe_ids, repeat(output_dir)),
                ),
                total=len(probe_ids),
                desc="Downloading IDAT files",
            )
        )


def unzip_and_remove_gz_files(directory, use_sentrix_id=False):
    """Function to unzip .gz files and remove the original .gz files."""
    for file_path in directory.glob("*.gz"):
        output_path = file_path.with_suffix("")
        if use_sentrix_id:
            sentrix_name = output_path.name.split("_", 1)[1]
            output_path = output_path.with_name(sentrix_name)
        with gzip.open(file_path, "rb") as fi, open(output_path, "wb") as fo:
            shutil.copyfileobj(fi, fo)
        file_path.unlink()


def setup_tutorial_files(analysis_dir, reference_dir):
    """Prepare the directory structure and files for the tutorial.

    This function sets up the necessary directory structure, processes the
    tutorial data, and downloads required IDAT files for both analysis and
    reference.

    Args:
        analysis_dir (str or Path): Path to the directory for storing analysis
            files.
        reference_dir (str or Path): Path to the directory for storing
            reference files.
    """
    control = "Control (muscle tissue)"
    tutorial_csv_path = get_resource_path("mepylome", "data/tutorial.csv.gz")

    analysis_dir = Path(analysis_dir)
    reference_dir = Path(reference_dir)
    analysis_dir.mkdir(parents=True, exist_ok=True)
    reference_dir.mkdir(parents=True, exist_ok=True)

    tutorial_df = pd.read_csv(tutorial_csv_path)
    tutorial_df.drop(columns=["Geo_File_ID"]).to_csv(
        analysis_dir / "annotation.csv", index=False
    )
    is_control = tutorial_df["Diagnosis"] == control

    def _missing_files(dir_path, geo_ids):
        missing_files = []
        for geo_id in geo_ids:
            sentrix_id = geo_id.split("_", 1)[1]
            grn_idat_file = dir_path / f"{sentrix_id}_Grn.idat"
            red_idat_file = dir_path / f"{sentrix_id}_Red.idat"
            if not grn_idat_file.exists() or not red_idat_file.exists():
                missing_files.append(geo_id)
        return missing_files

    missing_analysis_files = _missing_files(
        analysis_dir, tutorial_df["Geo_File_ID"]
    )
    if missing_analysis_files:
        download_geo_samples(analysis_dir, missing_analysis_files)
        unzip_and_remove_gz_files(analysis_dir, use_sentrix_id=True)

    missing_reference_files = _missing_files(
        reference_dir, tutorial_df[is_control]["Geo_File_ID"]
    )
    if missing_reference_files:
        download_geo_samples(reference_dir, missing_reference_files)
        unzip_and_remove_gz_files(reference_dir, use_sentrix_id=True)


def is_file_like(obj):
    """Check if the object is a file-like object.

    For objects to be considered file-like, they must be an iterator AND have
    either a `read` and/or `write` method as an attribute.  Note: file-like
    objects must be iterable, but iterable objects need not be file-like.

    Arguments:
        obj (any): The object to check.

    Returns:
        boolean: description

    Examples:
        >>> buffer(StringIO("data"))
        >>> is_file_like(buffer)
        True
        >>> is_file_like([1, 2, 3])
        False
    """
    return all(hasattr(obj, attr) for attr in ("read", "write", "__iter__"))


def get_file_object(filepath_or_buffer):
    """Returns a file-like object for the given input.

    Args:
        filepath_or_buffer (str or file-like object): The file path or
            file-like object to be processed. Can be a gz-archived file.

    Returns:
        file-like object: A file-like object for reading the file.
    """
    if is_file_like(filepath_or_buffer):
        return filepath_or_buffer

    if PurePath(filepath_or_buffer).suffix == ".gz":
        return gzip.open(filepath_or_buffer, "rb")

    return open(filepath_or_buffer, "rb")


def get_csv_file(file_or_archive, filename):
    """Retrieve a CSV file from a regular file or a ZIP archive.

    This function extracts a specific CSV file from either a regular CSV file
    or a ZIP archive and returns it as a file-like object.

    Examples:
        >>> get_csv_file('archive.zip', 'example.csv')
        >>> get_csv_file('/path/to/example.csv', 'example.csv')
    """
    file_or_archive = Path(file_or_archive)

    if file_or_archive.suffix == ".csv":
        return file_or_archive.open("rb")

    if file_or_archive.suffix == ".zip":
        with zipfile.ZipFile(file_or_archive, "r") as archive:
            file_list = archive.namelist()
            file_match = next(
                (f for f in file_list if f.endswith(filename)), None
            )
            if file_match:
                return archive.open(file_match, "r")
            msg = f"File '{filename}' not found in the ZIP archive."
            raise FileNotFoundError(msg)
    else:
        msg = "Unsupported file type. Only '.csv' and '.zip' are supported."
        raise ValueError(msg)


def reset_file(filepath_or_buffer):
    """Attempts to return the open file to the beginning if it is seekable."""
    if hasattr(filepath_or_buffer, "seek"):
        filepath_or_buffer.seek(0)



"""mepylome package.

This package provides tools for handling Illumina methylation arrays and
performing methylation analysis.
"""

import logging
import warnings

from mepylome.dtypes import (
    CNV,
    Annotation,
    ArrayType,
    IdatParser,
    Manifest,
    MethylData,
    RawData,
    ReferenceMethylData,
    clear_cache,
    idat_basepaths,
)
from mepylome.utils import make_log_file

# Suppress pyranges warnings
warnings.simplefilter(action="ignore", category=FutureWarning)

LOG_FILE = make_log_file("stdout")


def setup_logging():
    logger = logging.getLogger("mepylome")

    if logger.hasHandlers():
        logger.handlers.clear()

    logger.setLevel(logging.INFO)

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    # File handler
    file_handler = logging.FileHandler(LOG_FILE, mode="a")
    file_handler.setLevel(logging.INFO)

    # Formatter with no milliseconds
    log_format = "%(asctime)s [%(module)s] %(message)s"
    formatter = logging.Formatter(log_format, "%H:%M:%S")
    console_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)

    # Add handlers to logger
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    # Prevent root logger propagation to avoid Colab conflicts
    logger.propagate = False

    # Don't show logging statements of other libraries.
    logging.getLogger("werkzeug").setLevel(logging.ERROR)

    logger.debug("Logging is set up")


setup_logging()

__all__ = [
    "ArrayType",
    "Annotation",
    "CNV",
    "clear_cache",
    "IdatParser",
    "Manifest",
    "MethylData",
    "LOG_FILE",
    "RawData",
    "ReferenceMethylData",
    "idat_basepaths",
]

# If C++ parser should be added, before installation do: export MEPYLOME_CPP=1
try:
    from _mepylome import IdatParser as _IdatParser

    __all__ += ["_IdatParser"]
except Exception:
    pass



"""Processes all new IDAT files and generate beta values.

The generated beta files are compatible with the 450k array.

Author: Jon Brugger
Date: 2024-03
"""

import logging
import subprocess
import sys
from multiprocessing import Pool
from pathlib import Path

import numpy as np
from tqdm import tqdm

from mepylome import Manifest, MethylData, RawData
from mepylome.utils import CONFIG

INDEX_FILE = "/applications/reference_data/betaEPIC450Kmix_bin/index.csv"
NEW_IDAT_DIR = "/data/epidip_IDAT"
OUTPUT_DIR = "/data/epidip_CpGs_mepylome/"
BETA_BIN_FILE_SIZE = 3207696
ENDING_GRN = "_Grn.idat"
ENDING_BETAS = "_betas_filtered.bin"
LOG_FILE = "/applications/tmp/beta.logger"
ENDING_ERROR = CONFIG["suffixes"]["cnv_error"]

logging.basicConfig(
    format="%(asctime)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
    handlers=[
        logging.FileHandler(LOG_FILE, mode="w"),
        logging.StreamHandler(),
    ],
)
logger = logging.getLogger(__name__)
logger.info("Processing IDAT to beta values")


if not Path(OUTPUT_DIR).exists():
    Path(OUTPUT_DIR).mkdir(parents=True)


def result_filepath(filepath, error=False):
    ending = ENDING_ERROR if error else ENDING_BETAS
    return Path(OUTPUT_DIR, filepath.name + ending)


def write_error_file(filepath, error):
    idat_files = Path(NEW_IDAT_DIR, filepath.name)
    error_file = Path(OUTPUT_DIR, filepath.name + ENDING_ERROR)
    command = ["ls", "-lh", f"{idat_files}*"]
    files_on_disk = subprocess.check_output(command).decode("utf-8")
    error_message = (
        "During processing '"
        + filepath.name
        + "' the following error occurred:\n\n"
        + str(error)
        + "\n\nCorresponding files on disk:\n"
        + files_on_disk
        + "\n\n\nTo recalculate, delete this file."
    )
    with error_file.open("w") as f:
        f.write(error_message)


def process_idat_file(idat_basepath):
    filepath = result_filepath(idat_basepath)
    if Path(filepath).exists():
        if Path(filepath).stat().st_size == BETA_BIN_FILE_SIZE:
            logger.info("Betas allready caluclated: %s", filepath.name)
            return
        logger.info("Recalulate faulty betas: %s", filepath.name)
    raw = RawData(idat_basepath)
    methyl = MethylData(raw)
    betas_450k_df = methyl.betas_at(cpgs=cpg_index_450k, fill=0.49)
    betas_450k_np = betas_450k_df.iloc[:, 0].values
    betas_450k_np.tofile(filepath)


def process_idat_file_safe(idat_basepath):
    try:
        process_idat_file(idat_basepath)
    except Exception as e:
        write_error_file(idat_basepath, error=e)


with Path(INDEX_FILE).open() as f:
    cpg_index_450k = np.array(f.read().splitlines())


grn_idat_files = [
    x
    for x in list(Path(NEW_IDAT_DIR).iterdir())
    if str(x).endswith(ENDING_GRN)
]

basepaths = [
    Path(str(name).replace(ENDING_GRN, "")).expanduser()
    for name in grn_idat_files
]

old_basepaths = [
    path
    for path in basepaths
    if Path(result_filepath(path)).exists()
    or Path(result_filepath(path, error=True)).exists()
]

new_basepaths = list(set(basepaths).difference(old_basepaths))


def is_valid_file(filepath):
    return filepath.exists() and filepath.stat().st_size != BETA_BIN_FILE_SIZE


faulty_basepaths = [
    path for path in basepaths if is_valid_file(result_filepath(path))
]

if len(new_basepaths + faulty_basepaths) == 0:
    logger.info("No new or faulty IDAT files to process")
    sys.exit()

logger.info(
    "Processing %s new and %s incorrectly processed IDAT files.",
    len(new_basepaths),
    len(faulty_basepaths),
)

# Load manifests before multiprocessing
Manifest.load()


with Pool() as pool:
    betas_450k_results = list(
        tqdm(
            pool.imap(
                process_idat_file_safe, new_basepaths + faulty_basepaths
            ),
            total=len(new_basepaths),
        )
    )


logger.info("Processing IDAT to beta values completed")



# Tutorial (library)
# ==================


# This tutorial will guide you through the steps to analyze methylation data
# using the ``mepylome`` package. We'll cover setting up the environment,
# parsing IDAT files, working with manifest files, processing raw data,
# performing copy number variation (CNV) analysis, and conducting methylation
# analysis using UMAP plots with a GUI. You can find the code of this tutorial
# as a script in
# https://github.com/brj0/mepylome/blob/main/examples/rtd_tutorial.py.


# .. contents:: Contents of Tutorial
#    :depth: 3


# Setup
# -----


# First, import the required modules and set up the necessary directories:

from pathlib import Path

from mepylome import (
    CNV,
    IdatParser,
    Manifest,
    MethylData,
    RawData,
    ReferenceMethylData,
)

# Setup directories for the tutorial:
DIR = Path.home() / "mepylome" / "tutorial"
ANALYSIS_DIR = DIR / "tutorial_analysis"
REFERENCE_DIR = DIR / "tutorial_reference"

# Download and setup the necessary data files (approximately 653 MB):
from mepylome.utils import setup_tutorial_files

setup_tutorial_files(ANALYSIS_DIR, REFERENCE_DIR)


# Parsing IDAT files
# ------------------


# Define the IDAT file path:
idat_file = ANALYSIS_DIR / "200925700125_R07C01_Grn.idat"

# Parse the IDAT file:
idat_data = IdatParser(idat_file)

# Print overview of parsed data:
print(idat_data)
"""
IdatParser(
    file_size: 13686991
    num_fields: 19
    illumina_ids: array([ 1600101,  1600111, ..., 99810990, 99810992], dtype=int32)
    probe_means: array([15629,  8469, ...,  7971,   943], dtype=uint16)
    std_dev: array([1377,  408, ...,  702,  312], dtype=uint16)
    n_beads: array([16,  7, ...,  6, 10], dtype=uint8)
    mid_block: array([ 1600101,  1600111, ..., 99810990, 99810992], dtype=int32)
    red_green: 0
    mostly_null:
    barcode: 200925700125
    chip_type: BeadChip 8x5
    mostly_a: R07C01
    unknown_1:
    unknown_2:
    unknown_3:
    unknown_4:
    unknown_5:
    unknown_6:
    unknown_7:
)
"""

# The parsed data is available as attributes of the ``IdatParser`` object. For
# example the  Illumina IDs (probes IDs) can be accessed by:
ids = idat_data.illumina_ids

print(ids)
"""
[ 1600101  1600111  1600115 ... 99810978 99810990 99810992]
"""


# C++ Parser
# ~~~~~~~~~~


# If you installed mepylome with C++ support (see `installation
# <installation.html>`_) you can also use the C++ parser (input must be a
# string, not a Path object)
try:
    from mepylome import _IdatParser

    _idat_data = _IdatParser(str(idat_file))
    print("C++ parser available")

except ImportError:
    print("C++ parser NOT available")


# Manifest files
# --------------


# The mepylome package includes a ``Manifest`` class that provides
# functionality to download, process, and save Illumina manifest files
# internally in a efficient format (stored in ~/.mepylome). These manifest
# files contain information about the CpG sites on the methylation array,
# including genetic coordinates, probe types, and more.

# Load the available manifest files for different array types.
manifest_450k = Manifest("450k")
manifest_epic = Manifest("epic")
manifest_epic_v2 = Manifest("epicv2")
manifest_msa48 = Manifest("msa48")

# .. note::
#
#     The first time you run this, the manifest files will be downloaded and
#     saved locally to ~/.mepylome. This initial download might take some time.

# Obtain values from attributes:
probes_df = manifest_450k.data_frame
controls_df = manifest_450k.control_data_frame

# Print overview:
print(probes_df)
"""
            IlmnID  AddressA_ID  AddressB_ID  ...  N_CpG    End  Probe_Type
0       cg13869341     62703328     16661461  ...      2  15865           1
1       cg14008030     27651330           -1  ...      2  18827           2
2       cg12045430     25703424     34666387  ...      7  29407           1
3       cg20826792     61731400     14693326  ...      7  29425           1
4       cg00381604     26752380     50693408  ...      6  29435           1
...            ...          ...          ...  ...    ...    ...         ...
485572   rs1416770     28667385           -1  ...      0     -1           4
485573   rs1941955     33709340           -1  ...      0     -1           4
485574   rs2125573     25698376           -1  ...      0     -1           4
485575   rs2521373     12625304           -1  ...      0     -1           4
485576   rs4331560     10654345           -1  ...      0     -1           4

[485577 rows x 12 columns]
"""


# RawData
# -------


# The ``RawData`` class extracts both raw green and raw red signal intensity
# data from a IDAT file pair. You can initialize it using a base path to the
# IDAT files (without the _Grn.idat / _Red.idat suffix), or by providing the
# full path to either the Grn or Red IDAT file.
idat_file = ANALYSIS_DIR / "200925700125_R07C01_Red.idat"
## or
idat_file = ANALYSIS_DIR / "200925700125_R07C01_Grn.idat"
## or
idat_file = ANALYSIS_DIR / "200925700125_R07C01"
raw_data = RawData(idat_file)

# The data is saved within the following attributes:

## Intensity signals
raw_data.grn
raw_data.red

## Type of the array_type (e.g., 450k, EPIC)
raw_data.array_type

## Corresponding manifest file
raw_data.manifest

## IDs on the bead
raw_data.ids

# Print an overview of the raw data
print(raw_data)
"""
RawData():
**********

array_type: epic

manifest: epic

probes:
['200925700125_R07C01']

ids:
[ 1600101  1600111  1600115 ... 99810978 99810990 99810992]

_grn:
[[15629  8469  7015 ... 10228  7971   943]]

_red:
[[ 4429  1575 24955 ...  6594 15010  5336]]

grn:
          200925700125_R07C01
1600101                 15629
1600111                  8469
1600115                  7015
1600123                  7975
1600131                   938
...                       ...
99810958                 6292
99810970                  318
99810978                10228
99810990                 7971
99810992                  943

[1052641 rows x 1 columns]

red:
          200925700125_R07C01
1600101                  4429
1600111                  1575
1600115                 24955
1600123                 17707
1600131                  8967
...                       ...
99810958                 1881
99810970                 1936
99810978                 6594
99810990                15010
99810992                 5336

[1052641 rows x 1 columns]
"""

# RawData can also read multiple files of the same array type (used for
# reference files):
idat_file0 = ANALYSIS_DIR / "200925700125_R07C01_Grn.idat"
idat_file1 = ANALYSIS_DIR / "200925700133_R02C01"

raw_data_2 = RawData([idat_file0, idat_file1])

# Alternatively, read all IDAT files in a directory (supports recursive
# search):
raw_data_all = RawData(REFERENCE_DIR)


# MethylData
# ----------

# The ``MethylData`` class allows for processing raw intensity data and can
# calculate methylation signals as well as beta values. The raw data can be
# preprocessed using one of the following methods: 'illumina' (default),
# 'swan', or 'noob'. Initialize MethylData with raw data using the default
# 'illumina' preprocessing method.
methyl_data = MethylData(raw_data)

methyl_data_all = MethylData(raw_data_all)

# Alternatively, you can explicitly specify the 'illumina' preprocessing
# method.
methyl_data = MethylData(raw_data, prep="illumina")

# You can also initialize MethylData directly from an IDAT file path, without
# using ``RawData``. This is the preferred method if you want to obtain
# methylation signals or beta values.
methyl_data = MethylData(file=idat_file)

# Obtain various values via the attributes of the MethylData object:

## The methylation signals for the green and red channels.
methylated_signals = methyl_data.methylated
unmethylated_signals = methyl_data.unmethylated

## The corrected color signals.
corrected_green_signals = methyl_data.grn
corrected_red_signals = methyl_data.red

## The type of the array used (e.g., 450k, EPIC, EPICv2).
array_type = methyl_data.array_type

## The corresponding manifest file.
corresponding_manifest = methyl_data.manifest

# Print an overview of the methylation data.
print(methyl_data)
"""
MethylData():
*************

array_type: epic

manifest: epic

probes:
['200925700125_R07C01']

_grn:
[[16785.56811897  9044.70326442  7472.74551323 ... 10946.40456039
   8506.302329     908.14615612]]

_red:
[[ 3957.99684771  1303.20883282 23051.26201716 ...  5971.87773497
  13800.43272211  4801.6873626 ]]

grn:
          200925700125_R07C01
1600101          16785.568359
1600111           9044.703125
1600115           7472.745605
1600123           8510.626953
1600131            902.740540
...                       ...
99810958          6691.091309
99810970           232.442169
99810978         10946.404297
99810990          8506.302734
99810992           908.146179

[1052641 rows x 1 columns]

red:
          200925700125_R07C01
1600101           3957.996826
1600111           1303.208862
1600115          23051.261719
1600123          16309.179688
1600131           8179.240234
...                       ...
99810958          1587.849731
99810970          1639.010620
99810978          5971.877930
99810990         13800.432617
99810992          4801.687500

[1052641 rows x 1 columns]

methylated:
            200925700125_R07C01
IlmnID
cg14817997          2510.375488
cg26928153         10454.492188
cg16269199          7020.834473
cg13869341         30160.773438
cg14008030         19805.154297
...                         ...
cg10488260          2282.708496
cg14273923         12481.604492
cg09748881          6418.647461
cg07587934          9533.372070
cg16855331          5837.001465

[865859 rows x 1 columns]

unmethylated:
            200925700125_R07C01
IlmnID
cg14817997           855.783081
cg26928153           926.525330
cg16269199          3892.054932
cg13869341          5986.760742
cg14008030          8202.495117
...                         ...
cg10488260          8199.704102
cg14273923          2489.212646
cg09748881           950.663391
cg07587934          5264.926270
cg16855331         15618.041992

[865859 rows x 1 columns]
"""

# Beta values are a indicator wheather a CpG is methylated or not. They can be
# calculated for all sites of the corresponding array:
betas = methyl_data.betas

# You can also access the beta values at specific CpG sites (here at all the
# sites of the EPICv2 manifest). Missing data will be replaced with `fill`.
epicv2_cpgs = manifest_epic_v2.methylation_probes
beta_specific = methyl_data.betas_at(epicv2_cpgs, fill=0.5)


# Using Alternative Preprocessing Methods
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


# Preprocess the raw data using the SWAN method.
methyl_data_swan = MethylData(raw_data, prep="swan")

# Preprocess the raw data using the NOOB method.
methyl_data_noob = MethylData(raw_data, prep="noob")


# See `api <api.html>`_ for more information about SWAN and NOOB.


# Copy number variation (CNV)
# ---------------------------


# Copy number variations (CNV) are significant alterations in the genome
# involving the loss or gain of large DNA segments, often encompassing multiple
# genes. These variations are frequently linked to cancer development and can
# aid in tumor classification. The CNV profile can be calculated from signal
# intensity using methylation arrays. With the mepylome package, CNV can be
# efficiently calculated and visualized.


# **1. Set up analysis file**

# First, initialize your sample data for analysis:
sample = methyl_data


# **2. Set Up Reference Data**

# Within the reference directory there must be multiple CNV-neutral IDAT
# pairs of the **same array type** as `sample`.
reference = MethylData(file=REFERENCE_DIR)

# Alternatively, if the reference directory contains IDAT files of multiple
# array types, you can use ``ReferenceMethylData`` to load all files into
# memory. This way, the reference object can be used for multiple array types.
# The CNV class will automatically extract the files for the needed array type.
reference_all = ReferenceMethylData(REFERENCE_DIR)


# **3. Initialize CNV Analysis**

# Create an instance of the CNV class for the analysis, and fit the data (this
# is basically a linear regression model comparing `sample` signal with the
# `reference` signals at each CpG site):
cnv = CNV(sample, reference)

## Alternative with ReferenceMethylData
cnv = CNV(sample, reference_all)


# **4. Calculate CNV for Bins and Genes**

# Compute CNV values for genomic bins and gene regions:
cnv.set_bins()
cnv.set_detail()


# **5. Calculate CNV Segments**

# Use the binary circular segmentation algorithm for genome segmentation:
cnv.set_segments()

# .. note::
#
#     For this step, additional packages must be installed (see `installation
#     <installation.html>`_).


# **6. Streamlined Analysis**

# Alternatively, perform all CNV computations in a single call:
cnv = CNV.set_all(sample, reference)

## or
cnv = CNV.set_all(sample, reference_all)


# **7. Visualize CNV Data**

# Display an interactive plot using Plotly, where genes can be highlighted:
cnv.plot()


# Methylation Analysis
# --------------------


# **1. Set up analysis object and run GUI in browser**

# For methylation analysis, ensure you have the setup described in
# :ref:`general_setup`

# First, import the `MethylAnalysis` class from the `mepylome.analysis` module.
from mepylome.analysis import MethylAnalysis

# Create an instance of `MethylAnalysis` with the specified analysis and
# reference directories.
methyl_analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
)

# You can print an overview of the parameters of the object:
print(methyl_analysis)
"""
MethylAnalysis():
*****************

analysis_dir:
/home/username/mepylome/tutorial/tutorial_analysis

annotation:
/home/username/mepylome/tutorial/tutorial_analysis/annotation.csv

app:
None

betas_sel:
None

betas_all:
None

betas_dir:
/tmp/mepylome/analysis/betas-tutorial_analysis-illumina-3b616e0e24a8b0e2d443b777b8ad8b61

cnv_dir:
/tmp/mepylome/analysis/cnv-tutorial_analysis-tutorial_reference-illumina-False-f77099f7bcb04262a0456d122215ed4d

cnv_id:
None

cnv_plot:
Figure({
    'data': [], 'layout': {'template': '...', 'yaxis': {'range': [-2, 2...

cpg_selection:
top

cpgs:
['cg00000029' 'cg00000103' 'cg00000109' ... 'ch.X.97651759F'
 'ch.X.97737721F' 'ch.X.98007042R']

debug:
False

do_seg:
False

dropdown_id:
[]

host:
localhost

ids:
Index(['200925700133_R04C01', '201530470054_R05C01', '201869690203_R03C01',
       '201904410008_R05C01', '201869690168_R08C01', '201530470054_R01C01',
       '201869690203_R06C01', '200925700133_R03C01', '201904410008_R02C01',
       '201904410008_R04C01', '201530470054_R02C01', '201530470054_R03C01',
       '201904410008_R03C01', '201870610040_R03C01', '200925700133_R02C01',
       '200925700125_R07C01', '201530470054_R04C01', '201870610040_R04C01',
       '200925700133_R05C01', '201904410008_R06C01'],
      dtype='object')

ids_to_highlight:
None

load_full_betas:
False

n_cpgs:
25000

output_dir:
/tmp/mepylome/analysis

overlap:
False

port:
8050

precalculate_cnv:
False

prep:
illumina

raw_umap_plot:
Figure({
    'data': [{'customdata': array([['Chondrosarcoma', 'methylation clas...

reference_dir:
/home/username/mepylome/tutorial/tutorial_reference

selected_columns:
['Diagnosis']

umap_cpgs:
None

umap_df:
                       Umap_x    Umap_y  ...  Colour                 Umap_color
200925700133_R04C01  7.089545  1.715319  ...  2E3092  Osteosarcoma (high-grade)
201530470054_R05C01  8.902815  1.723695  ...  6B66AE              Osteoblastoma
201869690203_R03C01  3.801413  5.398326  ...  6282C2             Chondrosarcoma
201904410008_R05C01  6.164794  5.332215  ...  7C7E82    Control (muscle tissue)
201869690168_R08C01  3.486633  4.991620  ...  6282C2             Chondrosarcoma
201530470054_R01C01  8.228397  1.753262  ...  6B66AE              Osteoblastoma
201869690203_R06C01  3.823109  4.597386  ...  6282C2             Chondrosarcoma
200925700133_R03C01  7.502202  1.312939  ...  2E3092  Osteosarcoma (high-grade)
201904410008_R02C01  6.569446  5.207128  ...  7C7E82    Control (muscle tissue)
201904410008_R04C01  6.959955  4.955368  ...  7C7E82    Control (muscle tissue)
201530470054_R02C01  8.313415  2.253792  ...  6B66AE              Osteoblastoma
201530470054_R03C01  8.599755  1.344881  ...  6B66AE              Osteoblastoma
201904410008_R03C01  6.801136  5.651977  ...  7C7E82    Control (muscle tissue)
201870610040_R03C01  4.220203  5.398547  ...  6282C2             Chondrosarcoma
200925700133_R02C01  6.599361  2.232107  ...  2E3092  Osteosarcoma (high-grade)
200925700125_R07C01  6.635824  1.511574  ...  2E3092  Osteosarcoma (high-grade)
201530470054_R04C01  8.751877  2.230259  ...  6B66AE              Osteoblastoma
201870610040_R04C01  4.270664  4.871753  ...  6282C2             Chondrosarcoma
200925700133_R05C01  7.345413  2.310878  ...  2E3092  Osteosarcoma (high-grade)
201904410008_R06C01  6.281984  4.778615  ...  7C7E82    Control (muscle tissue)

[20 rows x 13 columns]

umap_dir:
/tmp/mepylome/analysis/umap-tutorial_analysis-illumina-25000-top-6ff1e8e7ec2b3a2f8d856ee634404085

umap_plot_path:
/tmp/mepylome/analysis/umap-tutorial_analysis-illumina-25000-top-6ff1e8e7ec2b3a2f8d856ee634404085/umap_plot.csv

upload_dir:
/tmp/mepylome/analysis/upload-tutorial_analysis-c8e2f8ac691e9c3deba2e880aa7c5251
"""

# To enable interactive analysis, you can launch a GUI. This will open a new
# tab in your web browser and you can start a GUI-based methylation analysis.
methyl_analysis.run_app(open_tab=True)


# `MethylAnalysis` has multiple parameters. For example, you can provide a
# custom set of CpG sites. For instance, you can set:
cpgs = Manifest("epic").methylation_probes[:10000]
methyl_analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
    cpgs=cpgs,
)

# Mepylome saves all results in a temporary directory ('/tmp/mepylome'). You
# can provide a custom directory for output:
OUTPUT_DIR = DIR / "output_dir"
OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
methyl_analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
    output_dir=OUTPUT_DIR,
)

# Here is a more comprehensive example with multiple custom parameters:
TEST_DIR = DIR / "test_dir"
methyl_analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
    output_dir=OUTPUT_DIR,
    ## New cases for validation, excluded from classifier training.
    test_dir=TEST_DIR,
    ## Load beta values for all CpG sites into memory
    load_full_betas=True,
    ## Use SWAN preprocessing method
    prep="swan",
    ## Provide annotation file (if not already in analysis_dir)
    annotation=ANALYSIS_DIR / "annotation.csv",
    ## Number of CpGs for UMAP analysis
    n_cpgs=5000,
    ## Analyze the 'best' CpG sites
    cpg_selection="top",
    ## Show segmentation intervals in CNV plot
    do_seg=True,
)

# Many parameters can be modified within the GUI application after
# initialization, but not all.


# **2. Set up beta values and generate UMAP**

# All calculations that can be performed within the GUI can also be done
# manually. For example, to extract the beta values:
methyl_analysis.set_betas()

# The beta values are then stored in:

methyl_analysis.betas_sel
"""
                     cg15836656  cg12823387  cg25563772  ...  cg22115994  cg27601809  cg00444740
200925700133_R04C01    0.117822    0.058700    0.058626  ...    0.295462    0.738677    0.474659
201530470054_R05C01    0.087871    0.050942    0.065696  ...    0.119668    0.305578    0.105583
201869690203_R03C01    0.810276    0.735967    0.721423  ...    0.751743    0.200545    0.757986
201904410008_R05C01    0.892641    0.777778    0.840475  ...    0.731927    0.636772    0.737969
201869690168_R08C01    0.811684    0.886028    0.888461  ...    0.725186    0.186773    0.812097
201530470054_R01C01    0.065221    0.050916    0.024315  ...    0.069606    0.408356    0.102764
201869690203_R06C01    0.948498    0.892334    0.859396  ...    0.776809    0.080326    0.875806
200925700133_R03C01    0.123428    0.193491    0.173582  ...    0.406198    0.754906    0.231158
201904410008_R02C01    0.868791    0.838671    0.922514  ...    0.704358    0.789430    0.617703
201904410008_R04C01    0.791002    0.818421    0.770333  ...    0.707137    0.851243    0.643282
201530470054_R02C01    0.058704    0.092474    0.122352  ...    0.173067    0.606287    0.340654
201530470054_R03C01    0.046851    0.043868    0.106606  ...    0.050913    0.281679    0.151072
201904410008_R03C01    0.718867    0.819853    0.899481  ...    0.701382    0.788357    0.659056
201870610040_R03C01    0.783809    0.905159    0.851122  ...    0.822050    0.104641    0.862166
200925700133_R02C01    0.054682    0.076460    0.089566  ...    0.087220    0.665932    0.159535
200925700125_R07C01    0.048604    0.093311    0.129750  ...    0.457459    0.867225    0.323574
201530470054_R04C01    0.048784    0.036592    0.048904  ...    0.105748    0.214249    0.056179
201870610040_R04C01    0.870172    0.896742    0.842515  ...    0.659235    0.134309    0.786843
200925700133_R05C01    0.048411    0.153956    0.164447  ...    0.462702    0.662564    0.290133
201904410008_R06C01    0.825491    0.828089    0.876774  ...    0.752596    0.833333    0.645860

[20 rows x 5000 columns]
"""

# To perform the UMAP algorithm:
methyl_analysis.compute_umap()

# The result of the UMAP algorithm is then stored in:
methyl_analysis.umap_df
"""
                       Umap_x    Umap_y  ...                    ID  Colour
200925700133_R04C01  6.644683 -7.024573  ...  REFERENCE_SAMPLE 367  2E3092
201530470054_R05C01  8.254219 -8.493701  ...  REFERENCE_SAMPLE 109  6B66AE
201869690203_R03C01  5.185830 -5.596291  ...   REFERENCE_SAMPLE 78  6282C2
201904410008_R05C01  7.058737 -5.482265  ...    REFERENCE_SAMPLE 2  7C7E82
201869690168_R08C01  4.727582 -5.662395  ...   REFERENCE_SAMPLE 80  6282C2
201530470054_R01C01  8.136439 -7.440626  ...  REFERENCE_SAMPLE 113  6B66AE
201869690203_R06C01  4.729284 -5.012035  ...   REFERENCE_SAMPLE 76  6282C2
200925700133_R03C01  6.980153 -7.641444  ...  REFERENCE_SAMPLE 368  2E3092
201904410008_R02C01  7.460121 -5.769195  ...    REFERENCE_SAMPLE 5  7C7E82
201904410008_R04C01  8.094703 -5.589474  ...    REFERENCE_SAMPLE 3  7C7E82
201530470054_R02C01  7.892371 -7.894600  ...  REFERENCE_SAMPLE 112  6B66AE
201530470054_R03C01  7.611932 -8.378901  ...  REFERENCE_SAMPLE 111  6B66AE
201904410008_R03C01  7.850667 -6.202358  ...    REFERENCE_SAMPLE 4  7C7E82
201870610040_R03C01  5.266904 -4.993761  ...   REFERENCE_SAMPLE 79  6282C2
200925700133_R02C01  6.583021 -7.879497  ...  REFERENCE_SAMPLE 369  2E3092
200925700125_R07C01  6.138983 -7.349401  ...  REFERENCE_SAMPLE 370  2E3092
201530470054_R04C01  8.566432 -7.849686  ...  REFERENCE_SAMPLE 110  6B66AE
201870610040_R04C01  5.656568 -5.504724  ...   REFERENCE_SAMPLE 77  6282C2
200925700133_R05C01  7.385612 -7.223170  ...  REFERENCE_SAMPLE 366  2E3092
201904410008_R06C01  7.607561 -5.099674  ...    REFERENCE_SAMPLE 1  7C7E82

[20 rows x 12 columns]
"""

# To generate the UMAP plot:
methyl_analysis.make_umap_plot()

# The Plotly object for the UMAP plot is then available in
# `methyl_analysis.umap_plot`.


# **3. CN-summary plots**

# Mepylome has the ability to calculate CN-summary plots, which provide an
# overview of copy number variations (CNV) across samples. To calculate these
# plots, you need to activate segmentation by setting 'do_seg' to True when
# initializing the analysis.
methyl_analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
    do_seg=True,
)

# The annotation dataframe contains metadata for each sample, including
# diagnosis information.
annotation_df = methyl_analysis.idat_handler.samples_annotated

# Loop through each unique diagnosis in the dataset to generate CN-summary
# plots for all diagnosis categories.
for diagnosis in annotation_df["Diagnosis"].unique():
    ## Filter sample IDs for the current diagnosis group.
    sample_ids = annotation_df[annotation_df["Diagnosis"] == diagnosis].index
    ## Generate CN-summary plot and retrieve data (if needed for further
    ## analysis).
    cn_plot, df_cn_summary = methyl_analysis.cn_summary(sample_ids)
    ## Update the plot layout with a title for the diagnosis group and label
    ## the y-axis.
    cn_plot.update_layout(
        title=f"CN-summary for {diagnosis}",
        yaxis_title="Proportion of CNV gains/losses",
    )
    ## Display the CN-summary plot in the browser.
    cn_plot.show()


# Supervised Classification
# -------------------------


# **1. Preimplemented Classifier**

# Methylome allows the use of preimplemented classifiers to predict methylation
# classes for samples. Below are examples of performing predictions on sample
# IDs and randomly generated values. Do setup as before:
import numpy as np
import pandas as pd
from sklearn.ensemble import (
    ExtraTreesClassifier,
    RandomForestClassifier,
)
from sklearn.feature_selection import (
    SelectKBest,
)
from sklearn.pipeline import Pipeline

from mepylome.analysis import MethylAnalysis

methyl_analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
)

# Perform prediction on sample IDs:

## Retrieve all sample IDs
ids = methyl_analysis.idat_handler.ids
## Set the annotation column for classification
methyl_analysis.idat_handler.selected_columns = ["Diagnosis"]
## Perform classification using a preimplemented classifier (pipeline with no
## scaler, SelectKBest as selector and RandomForestClassifier for
## classification) and cross-validation
clf_out = methyl_analysis.classify(ids=ids, clf_list="vtl-kbest-rf")

## Output accuracy scores:
print("Accuracy Scores:", clf_out[0].metrics["accuracy_scores"])
"""
Accuracy Scores: [1.0, 1.0, 1.0, 1.0, 1.0]
"""
## Detailed classifier report of first sample:
print(clf_out[0].reports["txt"][0])
"""
201530470054_R05C01
===================

Pipeline Structure:
- Scaler            : passthrough
- Feature_selection : SelectKBest
    k: 10000
- Classifier        : RandomForestClassifier

Metrics:
- Method    : 5-fold cross validation
- Samples   : 20
- Features  : 865859
- Accuracy  : 1.0000 (SD 0.0000)
- AUC       : 1.0000 (SD 0.0000)
- F1-Score  : 1.0000 (SD 0.0000)
- Precision : 1.0000 (SD 0.0000)
- Recall    : 1.0000 (SD 0.0000)

Classification Probability:
--------------------------------------
- Osteoblastoma             :  74.00 %
- Osteosarcoma (high-grade) :  23.00 %
- Control (muscle tissue)   :   2.00 %
- Chondrosarcoma            :   1.00 %
--------------------------------------
"""

# Perform prediction on beta values:

## Number of CpGs in the dataset
methyl_analysis.set_betas()
n_cpgs = methyl_analysis.betas_all.shape[1]
## Generate random beta values for 10 artificial samples
random_beta_values = pd.DataFrame(
    np.random.rand(10, n_cpgs), columns=methyl_analysis.betas_all.columns
)
## Perform classification on random values:
clf_out = methyl_analysis.classify(
    values=random_beta_values, clf_list="vtl-kbest-rf"
)

# **2. Train Your Own Classifiers**

# You can train custom classifiers using the scikit-learn API and incorporate
# them into the analysis. Below is an example using a Random Forest classifier.

## Set the annotation column for classification
methyl_analysis.idat_handler.selected_columns = ["Diagnosis"]

## Extract features (X) and target labels (y)
X = methyl_analysis.betas_all
y = methyl_analysis.idat_handler.features()

## Exclude test indices and invalid samples
valid_indices = [
    i
    for i, (idx, label) in enumerate(zip(X.index, y))
    if label and idx not in methyl_analysis.idat_handler.test_ids
]
X = X.iloc[valid_indices]
y = [y[i] for i in valid_indices]

## Choose classifier (maybe with tuned parameters)
rf_clf = RandomForestClassifier(
    n_estimators=500,
    max_depth=20,
    min_samples_split=5,
    min_samples_leaf=2,
    bootstrap=True,
    class_weight="balanced",
    random_state=42,
)

## Train the classifier
rf_clf.fit(X, y)

## If the classifier is already trained, mepylome will not repeat the training
## (and cross-validation) process
ids = methyl_analysis.idat_handler.ids
clf_out = methyl_analysis.classify(ids=ids, clf_list=rf_clf)

print(clf_out[0].reports["txt"][0])
"""
201530470054_R05C01
===================

Pipeline Structure:
- Classifier: RandomForestClassifier
    class_weight: balanced
    max_depth: 20
    min_samples_leaf: 2
    min_samples_split: 5
    n_estimators: 500
    random_state: 42

Metrics:
- Samples  : 20
- Features : 865859

Classification Probability:
--------------------------------------
- Osteoblastoma             :  83.19 %
- Osteosarcoma (high-grade) :   9.87 %
- Control (muscle tissue)   :   4.49 %
- Chondrosarcoma            :   2.45 %
--------------------------------------
"""

# **3. Use Untrained Scikit-Learn Classifiers**

# Untrained classifiers can also be passed directly for classification.
# Mepylome will then execute the training and cross-validation process. Below
# is an example using an Extra Trees classifier.

## Define an untrained Extra Trees classifier
et_clf = ExtraTreesClassifier(n_estimators=300, random_state=0)

## Perform classification
ids = methyl_analysis.idat_handler.ids
clf_out = methyl_analysis.classify(ids=ids, clf_list=et_clf)

print(clf_out[0].reports["txt"][0])
"""
201530470054_R05C01
===================

Pipeline Structure:
- Classifier: ExtraTreesClassifier
    n_estimators: 300
    random_state: 0

Metrics:
- Method    : 5-fold cross validation
- Samples   : 20
- Features  : 865859
- Accuracy  : 1.0000 (SD 0.0000)
- AUC       : 1.0000 (SD 0.0000)
- F1-Score  : 1.0000 (SD 0.0000)
- Precision : 1.0000 (SD 0.0000)
- Recall    : 1.0000 (SD 0.0000)

Classification Probability:
--------------------------------------
- Osteoblastoma             :  77.00 %
- Osteosarcoma (high-grade) :  13.00 %
- Control (muscle tissue)   :   6.00 %
- Chondrosarcoma            :   4.00 %
--------------------------------------
"""


# **4. Create a Custom Classifier**

# You can define a custom trained classifier by implementing a class that
# inherits from `TrainedClassifier`. This allows you to use non-standard models
# or apply additional functionality.

## Define a Custom Classifier
from mepylome.analysis.methyl_clf import TrainedClassifier


class CustomClassifier(TrainedClassifier):
    def __init__(self, clf):
        self.clf = clf
        self._classes = clf.classes_

    def predict_proba(self, betas, id_=None):
        return self.clf.predict_proba(betas)

    def classes(self):
        return self._classes

    def info(self, output_format="txt"):
        return "This text will be printed in reports."

    def metrics(self):
        return {"Key0": "Value0", "Key1": "Value1"}


## Initialize the custom classifier
custom_clf = CustomClassifier(rf_clf)

## Perform classification
clf_out = methyl_analysis.classify(ids=ids, clf_list=custom_clf)

print(clf_out[0].reports["txt"][0])
"""
201530470054_R05C01
===================

This text will be printed in reports.

Classification Probability:
--------------------------------------
- Osteoblastoma             :  83.19 %
- Osteosarcoma (high-grade) :   9.87 %
- Control (muscle tissue)   :   4.49 %
- Chondrosarcoma            :   2.45 %
--------------------------------------
"""


# **5. Add a Classifier into the MethylAnalysis Object**

# Finally, you can integrate a classifier directly into the `MethylAnalysis`
# object for seamless use in workflows and the interactive GUI app.

## Add a custom classifier into the MethylAnalysis object
pipeline = Pipeline(
    [
        ("feature_selection", SelectKBest(k=20000)),
        ("classifier", RandomForestClassifier(n_estimators=500)),
    ]
)
methyl_analysis.classifiers = {"model": pipeline, "name": "Custom RF"}

## Launch the interactive app with the integrated classifier. Now the classifier
## is available in the GUI under the name 'Custom RF'.
methyl_analysis.run_app(open_tab=True)


# For further details and advanced usage, refer to the mepylome documentation.



"""Command-line interface for running mepylome.

Usage:
    $ mepylome                     # Run mepylome with default settings
    $ mepylome -a /path/to/idats   # Specify analysis IDAT files directory
               -r /path/to/ref     # Specify reference IDAT directory
               -c 450k             # Specify CpG's to use
               -s                  # Add horizontal segmentation lines
    $ mepylome --help              # Show all parameters

"""

import argparse
import re
import textwrap
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path


def get_app_version():
    """Retrieve the app version from the package metadata."""
    try:
        return version("mepylome")
    except PackageNotFoundError:
        return "unknown"


def print_welcome_message():
    """Prints ASCII art welcome message."""
    app_version = get_app_version()
    welcome_message = rf"""
                             _
                            | |
  _ __ ___   ___ _ __  _   _| | ___  _ __ ___   ___
 | '_ ` _ \ / _ \ '_ \| | | | |/ _ \| '_ ` _ \ / _ \
 | | | | | |  __/ |_) | |_| | | (_) | | | | | |  __/
 |_| |_| |_|\___| .__/ \__, |_|\___/|_| |_| |_|\___|
                | |     __/ |
                |_|    |___/


Version: {app_version}
Starting Methylation Analysis...
"""
    print(welcome_message)


def absolute_path(path):
    """Converts a relative path to an absolute path."""
    return Path(path).absolute()


class SmartFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """Keeps new lines and doesn't break words, but still wraps lines.

    Source: https://gist.github.com/panzi/b4a51b3968f67b9ff4c99459fb9c5b3d
    """

    def _split_lines(self, text, width):
        lines = []
        for line in textwrap.dedent(text).strip().split("\n"):
            ident = re.match(r"^\s*", line).group(0)
            curr_line = [ident] if ident else []
            curr_len = len(ident)
            for word in line.split():
                if curr_line and curr_len + len(word) + 1 > width:
                    lines.append(" ".join(curr_line))
                    curr_line = [ident] if ident else []
                    curr_len = len(ident)
                curr_line.append(word)
                curr_len += len(word) + (1 if curr_line else 0)
            lines.append(" ".join(curr_line))
        return lines

    def _fill_text(self, text, width, indent):
        return "\n".join(
            indent + line
            for line in self._split_lines(text, width - len(indent))
        )

    def _format_action(self, action):
        """Adds an extra newline after each option."""
        return super()._format_action(action) + "\n"


def parse_args():
    """Parses command line arguments."""
    parser = argparse.ArgumentParser(
        description=(
            """
            Mepylome: Methylation Array Analysis Toolkit
            --------------------------------------------

            Command-line interface (CLI) for running the Mepylome methylation analysis GUI interface. This tool allows you to analyze methylation microarray data, visualize results, and run various analyses, including CNV (Copy Number Variation) and UMAP (Uniform Manifold Approximation and Projection) through a Dash-based GUI application.
            """
        ),
        epilog=(
            """
            Example usage:
            --------------

            1. Minimal Setup:
            Starts the GUI with only essential components. All required directories and settings must be provided within the GUI:

                mepylome

            2. Tutorial:
            Downloads a small IDAT dataset and launches the GUI with a preconfigured example analysis:

                mepylome --tutorial

            3. Standard Options (Recommended):
            Launches the GUI with the specified analysis and reference directories, with segmentation enabled:

                mepylome -a /path/to/analysis -r /path/to/references -s

            4. Multiple Options:
            Starts the GUI with additional parameters:

                python mepylome.py \\
                    --analysis_dir /path/to/analysis \\
                    --test_dir /path/to/test \\
                    --annotation /path/to/annotation.xlsx \\
                    --reference_dir /path/to/reference \\
                    --output_dir /path/to/output \\
                    --prep swan \\
                    --cpgs epic \\
                    --n_cpgs 10000 \\
                    --precalculate_cnv \\
                    --host 127.0.0.1 \\
                    --port 8050 \\
                    --debug
            """
        ),
        formatter_class=SmartFormatter,
    )
    parser.add_argument(
        "-a",
        "--analysis_dir",
        type=absolute_path,
        help="Path to the directory containing IDAT files for analysis.",
    )
    parser.add_argument(
        "-t",
        "--test_dir",
        type=absolute_path,
        help=(
            "Directory for test files, including new cases for analysis or "
            "validation. Files uploaded via the GUI will be placed here. If "
            "set to None, the application will automatically use a temporary "
            "directory."
        ),
    )
    parser.add_argument(
        "-A",
        "--annotation",
        type=str,
        help=(
            "Path to an annotation spreadsheet used to map sample files "
            "located in both `analysis_dir` and `test_dir`. One of the "
            "columns must contain the ID corresponding to the IDAT files "
            "(such as SentrixID or ID from files downloaded from GEO). If not "
            "provided, the system will attempt to identify the correct column "
            "automatically. If the annotation file is missing, it will search "
            "for a spreadsheet within the `analysis_dir` if available."
        ),
    )
    parser.add_argument(
        "-r",
        "--reference_dir",
        type=absolute_path,
        help=(
            "Directory containing CNV neutral reference IDAT files. Must be "
            "provided if you wanna generate CNV plots."
        ),
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=absolute_path,
        help=(
            "Path to the directory where output files will be saved. If not "
            "provided, the default directory '/tmp/mepylome/analysis' will be "
            "used."
        ),
    )
    parser.add_argument(
        "-p",
        "--prep",
        type=str,
        choices=["illumina", "swan", "noob"],
        default="illumina",
        help=(
            "Prepreparation method used for methylation microarrays: "
            "'illumina', 'swan', or 'noob'."
        ),
    )
    parser.add_argument(
        "-c",
        "--cpgs",
        type=str,
        default="auto",
        help=(
            """
            Specifies the CpG sites to analyze. Possible values:

                1. A path to a CSV file containing the CpG sites.

                2. A string specifying a predefined array type:

                    - '450k'   : The CpG sites from the Illumina 450k array.
                    - 'epic'   : The CpG sites from the Illumina EPIC array.
                    - 'epicv2' : The CpG sites from the Illumina EPIC v2 array.
                    - 'msa48'  : The CpG sites from the Illumina MSA array.

                3. A '+'-joined string of the options above combining multiple array types, returning the intersection of their CpG sites. For example:

                    - '450k+epic'  : CpG sites both in the 450k and EPIC arrays.
                    - 'epic+epicv2': CpG sites both in the EPIC and EPICv2 arrays.

                4. 'auto': Automatically detects all array types from IDAT files in analysis_dir and returns the intersection of CpG sites. This process may take longer as all files need to be read and, if necessary, decompressed.
            """
        ),
    )
    parser.add_argument(
        "-n",
        "--n_cpgs",
        default=25000,
        type=int,
        help="Number of CpG sites to select for UMAP.",
    )
    parser.add_argument(
        "-j",
        "--n_jobs_cnv",
        default=None,
        type=int,
        help=(
            "Number of parallel processes to use for CNV precalculation. If "
            "None, a reasonable number of cores will be automatically chosen "
            "based on the system and workload."
        ),
    )
    parser.add_argument(
        "-C",
        "--precalculate_cnv",
        action="store_true",
        help=(
            "If set, CNV data will be precalculated before the main analysis. "
            "This process takes approximately 2-5 seconds per case initially, "
            "but it will improve performance during runtime by reducing "
            "computation time."
        ),
    )
    parser.add_argument(
        "-l",
        "--no_load_full_betas",
        action="store_false",
        dest="load_full_betas",
        help=(
            "Prevent loading betas for all CpGs into memory setting "
            "`load_full_betas` to False. By default, betas are loaded to "
            "speed up the generation of sequential UMAP plots. Use this "
            "option only if you encounter memory overflow due to insufficient "
            "available memory (3-4 MB per sample)."
        ),
    )
    parser.add_argument(
        "-O",
        "--overlap",
        action="store_true",
        help=(
            "Only select IDAT files in the analysis directory that are also "
            "present in the annotation."
        ),
    )
    parser.add_argument(
        "-S",
        "--cpg_selection",
        type=str,
        choices=["top", "random"],
        default="top",
        help=(
            "Method to select CpG sites ('top' or 'random'). For 'top', CpG "
            "sites are selected based on their variation, taking the most "
            "varying ones. For 'random', CpG sites are randomly selected."
        ),
    )
    parser.add_argument(
        "-H",
        "--host",
        type=str,
        default="localhost",
        help="Host address for the Dash application.",
    )
    parser.add_argument(
        "-P",
        "--port",
        type=int,
        default=8050,
        help="Port number for the Dash application.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Flag to enable debug mode for the Dash application.",
    )
    parser.add_argument(
        "-s",
        "--do_seg",
        action="store_true",
        help=(
            "If set, enables segmentation analysis on CNV data and adds "
            "horizontal segmentation lines to the CNV plot. This will take an "
            "additional 2-5 seconds per sample."
        ),
    )
    parser.add_argument(
        "--tutorial",
        action="store_true",
        help=(
            "Downloads test IDAT files used in the tutorial and then launches "
            "the mepylome GUI session. Use this for a quick demonstration of "
            "how this package works."
        ),
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"{get_app_version()}",
    )

    return parser.parse_args()


def start_mepylome():
    """Entry point to start mepylome methylation analysis from command line."""
    args = parse_args()
    cli_args = {k: v for k, v in vars(args).items() if v is not None}

    print_welcome_message()

    if not cli_args["load_full_betas"] and cli_args["cpg_selection"] == "top":
        cli_args["cpg_selection"] = "random"
        msg = (
            "Warning: Since '-no_load_full_betas' is set, 'cpg_selection' has "
            "been automatically changed to 'random'.\n"
        )
        print(msg)

    from mepylome.analysis.methyl import MethylAnalysis

    if cli_args["tutorial"]:
        from mepylome.utils import setup_tutorial_files

        tutorial_dir = Path.home() / "mepylome" / "tutorial"
        cli_args["analysis_dir"] = tutorial_dir / "tutorial_analysis"
        cli_args["reference_dir"] = tutorial_dir / "tutorial_reference"
        if (
            not cli_args["analysis_dir"].exists()
            and not cli_args["reference_dir"].exists()
        ):
            # Download Tutorial IDAT files
            setup_tutorial_files(
                cli_args["analysis_dir"], cli_args["reference_dir"]
            )
        cli_args["load_full_betas"] = True
        cli_args["cpgs"] = "epic"

    cli_args.pop("tutorial", None)

    methyl_analysis = MethylAnalysis(**cli_args)
    methyl_analysis.run_app(open_tab=True)



"""Contains utilities to generate CNV plots."""

import io
import logging
import threading
import webbrowser
import zipfile
from functools import lru_cache
from pathlib import Path

import numpy as np
import pandas as pd
import plotly
import plotly.graph_objects as go

from mepylome.dtypes.cache import memoize
from mepylome.utils.files import ensure_directory_exists
from mepylome.utils.varia import CONFIG, MEPYLOME_TMP_DIR, get_free_port

logger = logging.getLogger(__name__)

PLOTLY_RENDER_MODE = "webgl"
ZIP_ENDING = CONFIG["suffixes"]["cnv_zip"]
THRESHOLD_BALANCED = 0.1
CHROMOSOME_DATA = CONFIG["genes"]["chromosome_data"]

ensure_directory_exists(MEPYLOME_TMP_DIR)
CNV_GRID = Path(MEPYLOME_TMP_DIR, f"cnv_grid_v{plotly.__version__}.json")


class Genome:
    """Data container for reference genome data."""

    def __init__(self):
        self.chrom = pd.DataFrame(CHROMOSOME_DATA)
        self.chrom["offset"] = [0] + np.cumsum(self.chrom["len"]).tolist()[:-1]
        self.chrom["center"] = self.chrom["offset"] + self.chrom["len"] // 2
        self.chrom["centromere_offset"] = (
            self.chrom["offset"]
            + (self.chrom["centromere_start"] + self.chrom["centromere_end"])
            // 2
        )
        self.length = (
            self.chrom["offset"].iloc[-1] + self.chrom["len"].iloc[-1]
        )

    def __iter__(self):
        """Enables looping over chromosomes."""
        return self.chrom.itertuples()

    def __len__(self):
        return self.length

    def __str__(self):
        """Prints overview of object for debugging purposes."""
        lines = [
            "Genome object:",
            f"length: {self.length}",
            f"chrom:\n{self.chrom}",
        ]
        return "\n".join(lines)

    def __repr__(self):
        return str(self)


reference_genome = Genome()

offset = {i.name: i.offset for i in reference_genome}


def get_df_from_zip(zip_file_path, extract=None):
    """Reads the CNV data from a zip file without extraxtion on disk."""
    extract = extract or ["bins", "detail", "segments"]
    zip_file_path = Path(zip_file_path).expanduser()
    zip_file = zip_file_path.name
    csv_file_templ = zip_file.replace(ZIP_ENDING, "_cnv_%s.csv")
    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
        extracted_dfs = []
        for filename in extract:
            try:
                with zip_ref.open(csv_file_templ % filename) as file_:
                    extracted_dfs.append(pd.read_csv(io.TextIOWrapper(file_)))
            except KeyError:
                extracted_dfs.append(None)
    return extracted_dfs


def add_offset(df, chrom_nm, col):
    """Calculates x-values of the CNV-plot."""
    return df[col] + df[chrom_nm].map(offset)


def get_x_mid(df):
    """Calculates the midpoint x-values of all chromosomes."""
    return df["Chromosome"].map(offset) + (df["Start"] + df["End"]) // 2


@lru_cache
def cnv_grid():
    """Returns chromosome grid for CNV Plot as a cached plotly object."""
    from plotly.io import from_json

    # Check if grid exists and return if available.
    if CNV_GRID.exists():
        with CNV_GRID.open() as f:
            return from_json(f.read())

    grid = go.Figure()
    grid.update_layout(
        coloraxis_showscale=False,
        xaxis={
            "linecolor": "black",
            "linewidth": 1,
            "mirror": True,
            "range": [0, len(reference_genome)],
            "showgrid": False,
            "ticklen": 10,
            "tickmode": "array",
            "ticks": "outside",
            "tickson": "boundaries",
            "ticktext": reference_genome.chrom.name,
            "tickvals": reference_genome.chrom.center,
            "zeroline": False,
        },
        yaxis={
            "linecolor": "black",
            "linewidth": 1,
            "mirror": True,
            "showline": True,
        },
        template="simple_white",
    )
    # Vertical line: centromere.
    for i in reference_genome.chrom.centromere_offset:
        grid.add_vline(x=i, line_color="black", line_dash="dot", line_width=1)
    # Vertical line: chromosomes.
    for i in [*reference_genome.chrom.offset.tolist(), len(reference_genome)]:
        grid.add_vline(x=i, line_color="black", line_width=1)
    # Draw horizontal line.
    grid.add_hline(y=0, line_color="black", line_width=1)
    # Save to disk
    grid.write_json(CNV_GRID)
    return grid


def cnv_bins_plot(data_frame, title, labels):
    """Create CNV plot from CNV data.

    Args:
        data_frame: DataFrame with columns 'x', 'y', and 'hover_data'
        title: Title of plot
        labels: Touple of labels for x-axis and y-axis.
    """
    scatter_trace = go.Scattergl(
        x=data_frame["x"],
        y=data_frame["y"],
        mode="markers",
        marker={
            "color": data_frame["y"],
            "colorscale": "RdBu_r",
            "cmin": -0.4,
            "cmax": 0.4,
        },
        text=data_frame["hover_data"],
        hovertemplate=(
            "<b>Value:</b> %{y}<br><b>Genes:</b> %{text}<extra></extra>"
        ),
        showlegend=False,
    )
    plot = go.Figure(cnv_grid())
    plot.add_trace(scatter_trace)
    plot.update_layout(
        title=title,
        yaxis_range=[-2, 2],
        xaxis_title=labels[0],
        yaxis_title=labels[1],
    )
    return plot


def add_segments(plot, seg_df):
    """Adds segments calculated by Circular Binary Segmentation (CBG)."""
    if seg_df is None:
        return plot
    seg_lines = [
        go.Scattergl(
            x=[seg["X_start"], seg["X_end"]],
            y=[seg["Median"], seg["Median"]],
            mode="lines",
            line={"color": "darkslategrey"},
            showlegend=False,
            name="",
        )
        for _, seg in seg_df.iterrows()
    ]
    plot.add_traces(seg_lines)
    return plot


def add_genes(plot, genes):
    """Add genes to the plot as bars with central crosses.

    Args:
        plot: The plot to which genes will be added.
        genes (list): A list of genes to be added to the plot.
    """
    # Draw NaN's with value 0
    genes["Median"] = genes["Median"].fillna(0)
    scatter_genes = go.Scattergl(
        customdata=genes[
            [
                "Name",  # 0
                "Chromosome",  # 1
                "Start",  # 2
                "End",  # 3
                "Len",  # 4
                "N_probes",  # 5
            ]
        ],
        hovertemplate=(
            "<b> %{customdata[0]} </b> "
            "%{customdata[1]}:%{customdata[2]:,}-%{customdata[3]:,} "
            "(hg19) <br>"
            "Value = %{y}<br>"
            "Gene length: %{customdata[4]:,} base pairs <br>"
            "Number or probes: %{customdata[5]} <br>"
        ),
        name="",
        marker_color="rgba(0,0,0,1)",
        mode="markers+text",
        marker_symbol="cross",
        textfont_color="rgba(0,0,0,1)",
        showlegend=False,
        text=genes.Name,
        textposition="top center",
        x=genes.X_mid,
        y=genes.Median,
        error_x={
            "type": "data",
            "symmetric": False,
            "array": genes.Len / 2,
            "arrayminus": genes.Len / 2,
        },
    )
    plot.add_trace(scatter_genes)
    return plot


def add_highlited_bins(plot, highlighted_bins):
    """Changes the color of the specified bins."""
    highlighted_bins_scatter = go.Scattergl(
        x=highlighted_bins.X_mid,
        y=highlighted_bins.Median,
        marker_color="magenta",
        mode="markers+text",
        showlegend=False,
    )
    plot.add_trace(highlighted_bins_scatter)
    return plot


@memoize
def find_genes_within_bins(bins, detail):
    """Determine genes overlapping with each bin and add them as a column.

    Args:
        bins (DataFrame): DataFrame containing bin information.
        detail (DataFrame): DataFrame containing gene information.

    Returns:
        Tuple(DataFrame, DataFrame): A tuple containing the bins DataFrame with
            assigned genes and the modified detail DataFrame.

    Raises:
        ValueError: If the 'bins' DataFrame is not sorted.
    """
    bins_X_start = add_offset(bins, "Chromosome", "Start")
    bins_X_end = add_offset(bins, "Chromosome", "End")

    if not bins_X_start.is_monotonic_increasing:
        msg = "'bins' is not sorted"
        raise ValueError(msg)

    detail_X_start = add_offset(detail, "Chromosome", "Start")
    detail_X_end = add_offset(detail, "Chromosome", "End")

    detail_Start_bin = np.digitize(detail_X_start, bins_X_start) - 1
    detail_Start_bin_ = np.digitize(detail_X_start, bins_X_end, right=True)
    detail_End_bin = np.digitize(detail_X_end, bins_X_start) - 1

    # Identify genes that start within gaps between bins
    detail_Start_bin[detail_Start_bin < detail_Start_bin_] += 1

    detail_Range = [
        np.arange(start, end)
        for start, end in zip(detail_Start_bin, detail_End_bin + 1)
    ]
    lengths = [len(x) for x in detail_Range]
    df_expanded = pd.DataFrame(
        {
            "Name": np.repeat(detail["Name"].values, lengths),
            "Range": np.concatenate(detail_Range),
        }
    )
    bins_Genes = (
        df_expanded
        # By sorting first, resultig string will be sorted too
        .sort_values(by="Name")
        .groupby("Range")["Name"]
        .agg(", ".join)
        .reindex(range(len(bins)), fill_value="")
        .reset_index(drop=True)
    )
    return bins_Genes, detail_Range


def _find_genes_within_bins(bins, detail):
    """Does (almost) the same as find_genes_within_bins (but 20x slower).

    Function is used for debugging as it is much easier to read.
    """
    bins["X_start"] = add_offset(bins, "Chromosome", "Start")
    bins["X_end"] = add_offset(bins, "Chromosome", "End")
    detail["X_start"] = add_offset(detail, "Chromosome", "Start")
    detail["X_end"] = add_offset(detail, "Chromosome", "End")
    bins["Genes"] = None
    bins["Genes"] = bins.apply(
        lambda bin_row: ", ".join(
            detail.loc[
                # The the intervals have a nonempty intersection
                ~(
                    (bin_row["X_end"] < detail["X_start"])
                    | (detail["X_end"] < bin_row["X_start"])
                ),
                "Name",
            ].sort_values()
        ),
        axis=1,
    )


def read_cnv_data_from_disk(cnv_dir, sample_id, extract=None):
    """Reads the components of a zip-file generated by CNV.write."""
    extract = extract or ["bins", "detail", "segments"]
    if isinstance(extract, str):
        extract = [extract]

    sample_zip = sample_id + ZIP_ENDING
    unzipped = get_df_from_zip(
        Path(cnv_dir) / sample_zip,
        extract=extract,
    )

    results = dict(zip(extract, unzipped))
    # Calculate some plot x-values
    if results.get("bins") is not None:
        results["bins"]["X_mid"] = get_x_mid(results["bins"])
    if results.get("detail") is not None:
        results["detail"]["X_mid"] = get_x_mid(results["detail"])
        results["detail"]["Len"] = (
            results["detail"]["End"] - results["detail"]["Start"]
        )
    if results.get("segments") is not None:
        results["segments"]["X_start"] = add_offset(
            results["segments"], "Chromosome", "Start"
        )
        results["segments"]["X_end"] = add_offset(
            results["segments"], "Chromosome", "End"
        )

    if len(results) == 1:
        return next(iter(results.values()))
    return tuple(results.values())


def cnv_plot_from_data(
    sample_id, bins, detail, segments, genes_fix, genes_sel
):
    """Generate a CNV plot from data calculated by the class CNV.

    Args:
        sample_id (str): The sample ID for the plot.
        bins (DataFrame): DataFrame containing bin information.
        detail (DataFrame): DataFrame containing gene information.
        segments (DataFrame): DataFrame containing segment information.
        genes_fix (list): List of genes to include in the plot.
        genes_sel (list): List of genes to include in the plot and highlight
            all associated bins.

    Returns:
        Plotly Figure: A Plotly figure representing the CNV plot.
    """
    logger.info("Make CNV plot: prepare data...")
    bins_Genes, detail_Range = find_genes_within_bins(
        bins[["Chromosome", "Start", "End"]],
        detail[["Chromosome", "Start", "End", "Name"]],
    )
    bins["Genes"] = bins_Genes
    detail["Range"] = detail_Range
    scatter_df = bins[["X_mid", "Median", "Genes"]]
    scatter_df.columns = ["x", "y", "hover_data"]

    # Base scatterplot
    logger.info("Make CNV plot: bins...")
    plot = cnv_bins_plot(
        data_frame=scatter_df,
        title=f"Sample ID: {sample_id}",
        labels=("", ""),
    )

    # Highlight bins adjacent to the added genes
    logger.info("Make CNV plot: genes...")
    genes_sel = genes_sel if genes_sel else []

    selected_genes_df = detail[detail["Name"].isin(genes_sel)]

    empty_range_genes = selected_genes_df["Name"][
        selected_genes_df["Range"].apply(lambda x: len(x) == 0)
    ].tolist()

    if empty_range_genes:
        msg = (
            "The following selected genes are outside of the CNV bins "
            "and therefore cannot be highlighted: "
            + ", ".join(empty_range_genes)
        )
        raise ValueError(msg)

    genes_x_range = selected_genes_df["Range"].explode().tolist()

    highlighted_bins = bins.loc[genes_x_range, ["X_mid", "Median"]]
    plot = add_highlited_bins(plot, highlighted_bins)

    # Add all added and important genes
    genes_to_plot = genes_fix + genes_sel
    gene_detail = detail[detail["Name"].isin(genes_to_plot)].copy()
    plot = add_genes(plot, gene_detail)

    # Draw the segments
    logger.info("Make CNV plot: segments...")
    plot = add_segments(plot, segments)
    return plot


class CNVPlot:
    """Creates an interactive plot for visualizing CNV data using Dash.

    This class sets up a Dash application to visualize copy number variation
    (CNV) data, including highlighting interactively genes.

    Attributes:
        cnv_dir (str): Path to the directory containing CNV data.
        cnv_file (str): Name of the CNV file in the directory.
        genes_fix (list): List of genes to highlight. Defaults to
            CONFIG["genes"]["default_genes_list"].
        host (str): Host address for running the Dash app. Defaults to
            "localhost".
        port (int): Port number for running the Dash app. Defaults to 8050.
        app (Dash): Dash application object created for the CNV plot.
    """

    def __init__(
        self,
        cnv_dir,
        cnv_file,
        genes=CONFIG["genes"]["default_genes_list"],
        host="localhost",
        port=8050,
    ):
        """Initializes a CNVPlot object.

        Args:
            cnv_dir (str): Path to the directory containing CNV data.
            cnv_file (str): Name of the CNV file in the directory.
            genes (list, optional): List of genes to highlight. Defaults to
                CONFIG["genes"]["default_genes_list"].
            host (str, optional): Host address for running the Dash app.
                Defaults to "localhost".
            port (int, optional): Port number for running the Dash app.
                Defaults to 8050.
        """
        self.cnv_dir = cnv_dir
        self.cnv_file = cnv_file
        self.genes_fix = genes
        self.host = host
        self.port = port
        self.app = self.get_app()

    def get_app(self):
        """Generates the dash app."""
        from dash import Dash, Input, Output, callback, dcc, html, no_update

        current_dir = Path(__file__).resolve().parent
        assets_folder = current_dir.parent / "data" / "assets"
        app = Dash(__name__, assets_folder=assets_folder)
        app._favicon = "favicon.svg"
        app.title = "mepylome"
        fig = go.Figure(layout=go.Layout(yaxis={"range": [-2, 2]}))
        app.layout = html.Div(
            [
                dcc.Store(id="memory"),
                dcc.Location(id="url", refresh=False),
                html.P(id="err", style={"color": "red"}),
                html.Div(id="page"),
                dcc.Dropdown(
                    options=[],
                    value=None,
                    placeholder="Select genes to highlight...",
                    id="dropdown",
                    multi=True,
                ),
                dcc.Graph(
                    id="graph",
                    figure=fig,
                    config={
                        "scrollZoom": False,
                        "doubleClick": "reset",
                        "modeBarButtonsToRemove": ["lasso2d", "select"],
                        "displaylogo": False,
                    },
                    style={"width": "80hh", "height": "90vh"},
                ),
            ]
        )

        @callback(
            [
                Output("graph", "figure"),
                Output("dropdown", "options"),
                Output("err", "children"),
            ],
            [
                Input("dropdown", "value"),
                Input("url", "pathname"),
            ],
        )
        def update_graph(genes_sel, url_path):
            sample_id = url_path[1:]
            try:
                bins, detail, segments = read_cnv_data_from_disk(
                    self.cnv_dir, sample_id
                )
            except FileNotFoundError:
                return no_update, no_update, f"Sample ID {sample_id} not found"
            plot = cnv_plot_from_data(
                sample_id,
                bins,
                detail,
                segments,
                self.genes_fix,
                genes_sel,
            )
            return plot, detail.sort_values(by="Name").Name, ""

        return app

    def run_app(self):
        """Opens new tab and runs app in browser."""
        init_sample_id = self.cnv_file.replace(ZIP_ENDING, "")
        free_port = get_free_port(self.port)

        def open_browser_tab():
            webbrowser.open_new_tab(
                f"http://{self.host}:{free_port}/{init_sample_id}"
            )

        threading.Timer(1, open_browser_tab).start()
        self.app.run(
            debug=True, host=self.host, use_reloader=False, port=free_port
        )


def _cn_summary_per_chrom(df_seg, threshold=0.1):
    """Generates a CN summary on disjoint intervals for a single chromosome."""
    df_seg["Gain"] = df_seg["Median"] > threshold
    df_seg["Loss"] = df_seg["Median"] < -threshold
    df_seg["Balanced"] = ~(df_seg["Gain"] | df_seg["Loss"])
    chromosome = df_seg.iloc[0]["Chromosome"]
    # Determine boundaries for disjoint segments
    boundaries = sorted(set(df_seg["Start"]).union(set(df_seg["End"])))
    result_intervals = []
    for i in range(1, len(boundaries)):
        start = boundaries[i - 1]
        end = boundaries[i]
        involved_segments = df_seg[
            (df_seg["Start"] <= start) & (df_seg["End"] >= end)
        ]
        gain_ratio = involved_segments["Gain"].mean()
        loss_ratio = involved_segments["Loss"].mean()
        result_intervals.append(
            {
                "Chromosome": chromosome,
                "Start": start,
                "End": end,
                "Gain_ratio": gain_ratio,
                "Loss_ratio": -loss_ratio,
                "Balanced_ratio": 1 - gain_ratio - loss_ratio,
            }
        )
    return pd.DataFrame(result_intervals)


def get_cn_summary(cnv_dir, sample_ids):
    """Generate a CNV summary plot for a given set of samples.

    This function reads CNV segment data from disk for a list of sample IDs,
    computes disjoint intervals for each chromosome, and generates a visual
    plot of gain/loss ratios across chromosomes.

    Args:
        cnv_dir (str or Path): The directory path where CNV data files are
            stored.
        sample_ids (list of str): A list of sample IDs to process.

    Returns:
        plot (plotly.graph_objects.Figure): A Plotly figure object containing
            the gain/loss ratio plot.
        df_cn_summary (pd.DataFrame): A pandas DataFrame containing the
            data used for the plot.
    """
    segment_list = []
    for sample_id in sample_ids:
        segment = read_cnv_data_from_disk(
            cnv_dir, sample_id, extract="segments"
        )
        if segment is None:
            msg = f"CNV for {sample_id} does not contain 'segments'."
            raise ValueError(msg)
        segment["Id"] = sample_id
        segment_list.append(segment)
    segments = pd.concat(segment_list)
    df_cn_summary = pd.concat(
        [
            _cn_summary_per_chrom(segments_on_chrom)
            for _, segments_on_chrom in segments.groupby("Chromosome")
        ]
    )
    df_cn_summary["X_start"] = add_offset(df_cn_summary, "Chromosome", "Start")
    df_cn_summary["X_end"] = add_offset(df_cn_summary, "Chromosome", "End")
    loss_color = "#1F77B4"
    gain_color = "#D62728"
    contour_color = "black"
    plot = go.Figure(cnv_grid())
    for chrom in df_cn_summary["Chromosome"].unique():
        # If there are gaps in the genome, ratios may be nan
        df_cn_summary_chrom = df_cn_summary[
            df_cn_summary["Chromosome"] == chrom
        ].fillna(0)
        first_start = df_cn_summary_chrom.iloc[0]["X_start"]
        x_vals_gain = [first_start]
        y_vals_gain = [0]
        x_vals_loss = [first_start]
        y_vals_loss = [0]
        for _, row in df_cn_summary_chrom.iterrows():
            x_vals_gain.extend([row["X_start"], row["X_end"]])
            y_vals_gain.extend([row["Gain_ratio"], row["Gain_ratio"]])
            x_vals_loss.extend([row["X_start"], row["X_end"]])
            y_vals_loss.extend([row["Loss_ratio"], row["Loss_ratio"]])
        last_end = df_cn_summary_chrom.iloc[-1]["X_end"]
        x_vals_gain.append(last_end)
        y_vals_gain.append(0)
        x_vals_loss.append(last_end)
        y_vals_loss.append(0)
        plot.add_trace(
            go.Scatter(
                x=x_vals_gain,
                y=y_vals_gain,
                mode="lines",
                fill="tozeroy",
                fillcolor=gain_color,
                line={"color": contour_color, "width": 1.5},
                hovertemplate="<b>Value:</b> %{y}<br><extra></extra>",
                showlegend=False,
            )
        )
        plot.add_trace(
            go.Scatter(
                x=x_vals_loss,
                y=y_vals_loss,
                mode="lines",
                fill="tozeroy",
                fillcolor=loss_color,
                line={"color": contour_color, "width": 1.5},
                hovertemplate="<b>Value:</b> %{y}<br><extra></extra>",
                showlegend=False,
            )
        )
    plot.update_layout(
        yaxis_range=[-1, 1],
    )
    return plot, df_cn_summary



"""Contains methods for supervised learning.

Non supervised classifiers (random forest, k-nearest neighbors, neural
networks) for predicting the methylation class.
"""

import ast
import hashlib
import logging
import pickle
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import (
    LinearDiscriminantAnalysis,
    QuadraticDiscriminantAnalysis,
)
from sklearn.ensemble import (
    AdaBoostClassifier,
    BaggingClassifier,
    ExtraTreesClassifier,
    GradientBoostingClassifier,
    HistGradientBoostingClassifier,
    RandomForestClassifier,
)
from sklearn.feature_selection import (
    SelectKBest,
)
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import (
    LogisticRegression,
    Perceptron,
    RidgeClassifier,
    SGDClassifier,
)
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
)
from sklearn.model_selection import (
    cross_val_predict,
)
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (
    MinMaxScaler,
    PowerTransformer,
    QuantileTransformer,
    RobustScaler,
    StandardScaler,
)
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

logger = logging.getLogger(__name__)


class TrainedClassifier(ABC):
    """Abstract base class for a trained classifier."""

    @abstractmethod
    def predict_proba(self, betas, id_=None):
        """Predicts the probability of the given input samples (betas).

        Args:
            betas (array-like): Input samples to predict probabilities.
            id_ (string, optional): Identifier of the input samples.

        Returns:
            array-like: Predicted probabilities for each sample.
        """

    @abstractmethod
    def classes(self):
        """Returns the list of classes for which predictions can be made.

        Returns:
            array-like: The classes the classifier can predict.
        """

    def info(self, output_format="txt"):
        """Provides additional information of the classifier or its pipeline.

        The info will be printed on the top of the classifiers report. Should
        contain classifier name and classifier metrics.

        Returns:
            str: A description of the classifier or its components.
        """
        return ""

    def model(self):
        """Returns the classifier model (usually sklearn object)."""
        return

    def metrics(self):
        """Returns the metric statistics (usually sklearn object)."""
        return {}

    def __str__(self):
        hash_str = hashlib.blake2b(
            self.info().encode(), digest_size=16
        ).hexdigest()
        return f"TrainedClassifier_{hash_str}"

    def __repr__(self):
        return str(self)


def _get_pipeline_description(clf, output_format="txt"):
    """Generates a detailed summary string of the pipeline structure."""

    def format_non_default_params(step):
        """Formats non-default parameters for a given step or classifier."""
        default_params = step.__class__().get_params()
        current_params = step.get_params()
        non_default_params = {
            k: v
            for k, v in current_params.items()
            if default_params.get(k) != v
        }
        if output_format == "html":
            return [
                (
                    f"<tr><td class='clf-label'>- {param}</td>"
                    f"<td class='clf-value'>{value}</td></tr>"
                )
                for param, value in non_default_params.items()
            ]
        return [
            f"- {param}: {value}"
            for param, value in non_default_params.items()
        ]

    if output_format == "html":
        lines = ["<h2>Classifier Structure</h2>", "<table>"]
    else:
        lines = ["Classifier Structure:"]
    if hasattr(clf, "steps"):
        step_name_len = max(len(name) for name, _ in clf.steps)
        for name, step in clf.steps:
            step_name = name.capitalize()
            step_type = (
                step if isinstance(step, str) else step.__class__.__name__
            )
            if output_format == "html":
                lines.append(
                    f"<tr><td class='clf-label'>{step_name}</td>"
                    f"<td class='clf-value'>{step_type}</td></tr>"
                )
            else:
                lines.append(f"{step_name:<{step_name_len}} : {step_type}")
            if not isinstance(step, str):
                lines.extend(format_non_default_params(step))
    else:
        clf_name = clf.__class__.__name__
        if output_format == "html":
            lines.append(
                f"<tr><td class='clf-label'>Classifier</td>"
                f"<td class='clf-value'>{clf_name}</td></tr>"
            )
        else:
            lines.append(f"Classifier: {clf_name}")
        lines.extend(format_non_default_params(clf))

    if output_format == "html":
        lines.append("</table>")

    return "\n".join(lines)


def _format_metrics(metrics):
    """Transforms classifier metrics to readable text."""

    def format_metric_scores(scores):
        if isinstance(scores, (list, np.ndarray)):
            return f"{np.mean(scores):.4f} (SD {np.std(scores):.4f})"
        return str(scores)

    formatted_stats = {}

    metric_keys = {
        "Method": "cv",
        "Samples": "n_samples",
        "Features": "n_features",
        "Accuracy": "accuracy_scores",
        "AUC": "auc_scores",
        "F1-Score": "f1_scores",
        "Precision": "precision_scores",
        "Recall": "recall_scores",
    }

    for key, metric in metric_keys.items():
        if metric in metrics:
            if metric == "cv":
                formatted_stats[key] = (
                    f"{metrics[metric].n_splits}-fold cross validation"
                )
            else:
                formatted_stats[key] = format_metric_scores(metrics[metric])

    return formatted_stats


class TrainedSklearnClassifier(TrainedClassifier):
    """Trained classifier implementation using fitted scikit-learn objecs."""

    def __init__(self, clf, X=None, metrics=None):
        self.clf = clf
        self._classes = clf.classes_
        self.ids = [] if X is None else X.index
        self._metrics = metrics or {
            "n_features": X.shape[1] if X is not None else 0,
            "n_samples": X.shape[0] if X is not None else 0,
        }

    def predict_proba(self, betas, id_=None):
        return self.clf.predict_proba(betas)

    def classes(self):
        return self._classes

    def info(self, output_format="txt"):
        description = _get_pipeline_description(self.clf, output_format)
        formatted_stats = _format_metrics(self._metrics)

        if output_format == "html":
            result = [f"{description}"]
            result.append("<h2>Metrics</h2>")
            result.append("<table>")
            for key, value in formatted_stats.items():
                result.append(
                    f"<tr><td class='metrics-label'>{key}</td>"
                    f"<td class='metrics-value'>{value}</td></tr>"
                )
            result.append("</table>")
        else:
            max_key_length = max(len(key) for key in formatted_stats)
            result = [description + "\n", "Metrics:"]
            for key, value in formatted_stats.items():
                result.append(f"{key:<{max_key_length}} : {value}")

        return "\n".join(result)

    def metrics(self):
        return self._metrics

    def model(self):
        return self.clf


class TrainedSklearnCVClassifier(TrainedClassifier):
    """A trained sklearn classifier with support for cross-validation.

    This class allows the use of precomputed cross-validation probabilities or
    standard predictions from a trained classifier for a given sample. It also
    provides additional statistics about the classifier.
    """

    def __init__(self, clf, probabilities_cv, X, metrics=None):
        self.clf = clf
        self.probabilities_cv = pd.DataFrame(probabilities_cv, index=X.index)
        self.ids = X.index
        self._metrics = metrics or {}
        self._classes = clf.classes_

    def predict_proba(self, betas, id_=None):
        if id_ is not None:
            id_ = np.ravel(id_)
            probabilities = np.zeros((len(betas), len(self._classes)))

            if len(betas) != len(id_):
                msg = "Length of 'betas' and 'id_' must match."
                raise ValueError(msg)

            mask_known_ids = np.isin(id_, self.ids)
            if any(mask_known_ids):
                probabilities[mask_known_ids] = self.probabilities_cv.loc[
                    id_[mask_known_ids]
                ].values
            unknown_betas = betas[~mask_known_ids]
            if len(unknown_betas) > 0:
                probabilities[~mask_known_ids] = self.clf.predict_proba(
                    unknown_betas
                )
            return probabilities

        return self.clf.predict_proba(betas)

    def classes(self):
        return self._classes

    def info(self, output_format="txt"):
        description = _get_pipeline_description(self.clf, output_format)
        formatted_stats = _format_metrics(self._metrics)

        if output_format == "html":
            result = [f"{description}"]
            result.append("<h2>Metrics</h2>")
            result.append("<table>")
            for key, value in formatted_stats.items():
                result.append(
                    f"<tr><td class='metrics-label'>{key}</td>"
                    f"<td class='metrics-value'>{value}</td></tr>"
                )
            result.append("</table>")
        else:
            max_key_length = max(len(key) for key in formatted_stats)
            result = [description + "\n", "Metrics:"]
            for key, value in formatted_stats.items():
                result.append(f"{key:<{max_key_length}} : {value}")

        return "\n".join(result)

    def metrics(self):
        return self._metrics

    def model(self):
        return self.clf


class VarianceThresholdLite(BaseEstimator, TransformerMixin):
    """low memory version of sklearn.feature_selection.VarianceThreshold."""

    def __init__(self, threshold=1e-4):
        self.threshold = threshold

    def fit(self, X, y=None):
        X_array = X.values if isinstance(X, pd.DataFrame) else np.asarray(X)
        if X_array.ndim != 2:
            msg = "Input X must be 2D."
            raise ValueError(msg)
        self.variances_ = np.var(X_array, axis=0)
        self.support_mask_ = self.variances_ > self.threshold
        if not np.any(self.support_mask_):
            msg = f"No features meet the variance threshold {self.threshold}"
            raise ValueError(msg)
        return self

    def transform(self, X):
        X_array = X.values if isinstance(X, pd.DataFrame) else np.asarray(X)
        return X_array[:, self.support_mask_]

    def _get_support_mask(self):
        return self.variances_ > self.threshold


class TopVarianceSelector(BaseEstimator, TransformerMixin):
    """Low-memory selector for top-N features by variance."""

    def __init__(self, n_top=30000):
        self.n_top = n_top

    def fit(self, X, y=None):
        X_array = X.values if isinstance(X, pd.DataFrame) else np.asarray(X)
        if X_array.ndim != 2:
            msg = "Input X must be 2D."
            raise ValueError(msg)

        # Compute variance per column
        self.variances_ = np.var(X_array, axis=0)

        # Get indices of top N variances
        if self.n_top > X_array.shape[1]:
            msg = (
                f"n_top={self.n_top} exceeds number of features "
                "{X_array.shape[1]}"
            )
            raise ValueError(msg)
        self.top_indices_ = np.argsort(self.variances_)[-self.n_top :]

        return self

    def transform(self, X):
        X_array = X.values if isinstance(X, pd.DataFrame) else np.asarray(X)
        return X_array[:, self.top_indices_]

    def get_support(self, indices=False):
        if indices:
            return self.top_indices_
        mask = np.zeros_like(self.variances_, dtype=bool)
        mask[self.top_indices_] = True
        return mask


def parse_component_key(key):
    """Parse component name and kwargs like 'lr(n_iter=200)'."""
    match = re.match(r"^(\w+)(?:\((.*)\))?$", key)
    if not match:
        raise ValueError(f"Invalid component key format: {key}")
    name, kwargs_str = match.groups()
    args = []
    kwargs = {}
    if kwargs_str:
        tree = ast.parse(f"f({kwargs_str})", mode="eval")
        if not isinstance(tree.body, ast.Call):
            raise ValueError(f"Malformed kwargs in: {key}")
        args = [ast.literal_eval(arg) for arg in tree.body.args]
        kwargs = {
            kw.arg: ast.literal_eval(kw.value) for kw in tree.body.keywords
        }
    return name, args, kwargs


def make_clf_pipeline(step_keys, X_shape=None, cv=None):
    """Sklearn pipeline with scaling, feature selection, and classifier."""
    scalers = {
        "minmax": MinMaxScaler,
        "power": PowerTransformer,
        "quantile": QuantileTransformer,
        "robust": RobustScaler,
        "std": StandardScaler,
    }
    selectors = {
        "kbest": lambda **kwargs: SelectKBest(**{"k": 10000, **kwargs}),
        "pca": PCA,
        "top": TopVarianceSelector,
        "vtl": VarianceThresholdLite,
    }
    if X_shape and cv:
        n_splits = cv if isinstance(cv, int) else cv.n_splits
        n_components_pca = min(((n_splits - 1) * X_shape[0] // n_splits), 50)
        selectors["pca_auto"] = lambda **kwargs: PCA(
            **{"n_components": n_components_pca, **kwargs}
        )
    models = {
        "ada": AdaBoostClassifier,
        "bag": BaggingClassifier,
        "dt": DecisionTreeClassifier,
        "et": ExtraTreesClassifier,
        "gb": GradientBoostingClassifier,
        "gp": GaussianProcessClassifier,
        "hgb": HistGradientBoostingClassifier,
        "knn": KNeighborsClassifier,
        "lda": LinearDiscriminantAnalysis,
        "lr": LogisticRegression,
        "mlp": lambda **kwargs: MLPClassifier(**{"verbose": True, **kwargs}),
        "nb": GaussianNB,
        "perceptron": Perceptron,
        "qda": QuadraticDiscriminantAnalysis,
        "rf": RandomForestClassifier,
        "ridge": RidgeClassifier,
        "sgd": SGDClassifier,
        "svc": lambda **kwargs: SVC(
            **{"probability": True, "verbose": True, **kwargs}
        ),
    }
    components = {
        **{key: ("scaler", val) for key, val in scalers.items()},
        **{key: ("feature_selector", val) for key, val in selectors.items()},
        **{key: ("classifier", val) for key, val in models.items()},
    }
    components_full_name = {
        clf().__class__.__name__: (name, clf)
        for name, clf in components.values()
    }
    components.update(components_full_name)
    if isinstance(step_keys, str):
        step_keys = step_keys.split("-")
    pipeline_components = []
    for i, full_key in enumerate(step_keys):
        base_key, args, kwargs = parse_component_key(full_key)
        if base_key not in components:
            raise ValueError(f"Invalid step key: {base_key}")
        type_name, constructor = components[base_key]
        component = constructor(*args, **kwargs)
        step_name = f"{i + 1}-{type_name}"
        pipeline_components.append((step_name, component))

    return Pipeline(pipeline_components)


def make_reports(prediction, info, output_format="txt"):
    """Generates detailed reports from classifier predictions.

    Args:
        prediction (pd.DataFrame): DataFrame containing predicted probabilities
            for each class, indexed by sample IDs.
        info (str): A description of the classifier such as its name and
            metrics. Will be printed before predictions.
        output_format (str): The format of the report ('txt' or 'html').
            Defaults to 'txt'.

    Returns:
        list[str]: A list of detailed string reports, one for each sample.
    """
    reports = []
    for sample_id, row in prediction.iterrows():
        top_predictions = row[row > 0].nlargest(10) * 100
        if output_format == "txt":
            report_lines = [
                str(sample_id),
                "=" * len(str(sample_id)),
                f"\n{info}\n",
                "Classification Probability:",
            ]
            label_len = max(len(str(label)) for label in top_predictions.index)
            formatted_lines = [
                f"{label:<{label_len}} : {probability:6.2f} %"
                for label, probability in top_predictions.items()
            ]
            n_dashes = (
                max(len(line) for line in formatted_lines)
                if formatted_lines
                else 0
            )
            dashes = "-" * n_dashes
            report_lines.extend([dashes, *formatted_lines, dashes])
        if output_format == "html":
            report_lines = [
                f"<h1>{sample_id}</h1>",
                f"{info}",
                "<h2>Classification Probability</h2>",
                "<div class='classification-result'>",
                "<table>",
            ]
            report_lines.extend(
                (
                    f"<tr><td class='prob-label'>{label}</td>"
                    f"<td class='prob-value'>{probability:.2f}%</td></tr>"
                )
                for label, probability in top_predictions.items()
            )
            report_lines.extend(["</table>", "</div>"])
        reports.append("\n".join(report_lines))
    return reports


def make_classifier_report_page(reports, title=None):
    """Generates an HTML page for multiple classifier reports.

    Args:
        report_sections (list of str): A list of HTML reports obtained from
            'MethylAnalysis.classify'.
        itle (str): Title of the report page. Defaults to 'None'.

    Returns:
        str: A formatted HTML string.
    """
    title_str = f"<h1>{title}</h1><hr>" if title else ""
    body_string = "\n<hr>\n".join(reports)
    return f"""<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Classifier Report</title>
    <style>
        body {{
            font-family: Arial, sans-serif;
            background: #f4f7f9;
            color: #333;
            margin: 0 auto;
            max-width: 700px;
            line-height: 1.4;
        }}
        h1 {{
            text-align: center;
            margin-top: 10px;
            font-size: 1.5em;
        }}
        h2 {{
            font-size: 1.2em;
            margin-bottom: 5px;
        }}
        table {{
            width: 100%;
            max-width: 500px;
            margin: 10px auto;
            border-collapse: collapse;
        }}
        hr {{
            margin: 20px 0;
            border: none;
            height: 2px;
            background-color: #ccc;
        }}
        th, td {{
            padding: 3px 8px;
            text-align: left;
            border-bottom: 1px solid #ddd;
            font-size: 0.9em;
        }}
        td.clf-label, td.metrics-label, td.prob-label {{
            font-weight: bold;
        }}
        td.clf-value, td.metrics-value, td.prob-value {{
            text-align: right;
        }}
        .classification-result td {{
            background: #ffebe6;
        }}
        .classification-result-k {{
            background: #ffe4b5;
            font-weight: bold;
            color: #d85c5c;
        }}
        @media (max-width: 768px) {{
            body {{
                max-width: 95%;
            }}
            table {{
                max-width: 100%;
            }}
        }}
    </style>
</head>
<body>
{title_str}
{body_string}
</body>
</html>"""


def _is_ovr_or_ovo(clf):
    """Check if classifier is using One-vs-Rest (OvR) or One-vs-One (OvO)."""
    if isinstance(clf, OneVsRestClassifier):
        return "ovr"
    if isinstance(clf, OneVsOneClassifier):
        return "ovo"
    if hasattr(clf, "decision_function_shape"):
        if clf.decision_function_shape == "ovr":
            return "ovr"
        if clf.decision_function_shape == "ovo":
            return "ovo"
    # Default for most classifiers
    return "ovr"


def cross_val_metrics(clf, X, y, probabilities_cv, cv):
    """Calculates cross-validation statistics for a classifier."""
    y_pred_cv = np.array(
        [clf.classes_[i] for i in np.argmax(probabilities_cv, axis=1)]
    )
    accuracy_scores = []
    auc_scores = []
    f1_scores = []
    multi_class_strategy = _is_ovr_or_ovo(clf)
    precision_scores = []
    recall_scores = []

    for _, test_idx in cv.split(X, y):
        y_true = np.array(y)[test_idx]
        probabilities_cv_test = probabilities_cv[test_idx, :]
        y_pred = y_pred_cv[test_idx]

        accuracy_scores.append(accuracy_score(y_true, y_pred))
        f1_scores.append(f1_score(y_true, y_pred, average="weighted"))
        precision_scores.append(
            precision_score(
                y_true, y_pred, average="weighted", zero_division=0
            )
        )
        recall_scores.append(
            recall_score(y_true, y_pred, average="weighted", zero_division=0)
        )

        # Calculate ROC AUC score
        if len(np.unique(y_true)) == 2:
            # Binary classification: use probabilities for the positive class
            roc_auc = roc_auc_score(y_true, probabilities_cv_test[:, 1])
        else:
            # Multiclass classification: use the full probabilities matrix
            roc_auc = roc_auc_score(
                y_true,
                probabilities_cv_test,
                multi_class=multi_class_strategy,
                average="weighted",
            )
        auc_scores.append(roc_auc)

    return {
        "cv": cv,
        "n_samples": X.shape[0],
        "n_features": X.shape[1],
        "accuracy_scores": accuracy_scores,
        "auc_scores": auc_scores,
        "f1_scores": f1_scores,
        "precision_scores": precision_scores,
        "recall_scores": recall_scores,
    }


def is_trained(clf):
    """Checks if the given classifier has already been trained."""
    clf_ = clf
    if hasattr(clf, "steps"):
        clf_ = clf.steps[-1][1]
    trained_attributes = ["classes_"]
    return all(hasattr(clf_, attr) for attr in trained_attributes)


def train_clf(clf, X, y, save_path, cv, n_jobs=1):
    """Trains a classifier and stores the trained model to disk.

    If the classifier has already been trained (and saved), it loads the
    trained model. If not, it trains the classifier using the provided data and
    saves it to disk.

    Args:
        clf (classifier): The classifier to train or load.
        X (array-like): The feature matrix.
        y (array-like): The target labels.
        save_path (Path): The path where the trained model should be
            saved.
        cv (int or cross-validation generator): Determines the cross-validation
            splitting strategy.
        n_jobs (int): Number of parallel processes to run.

    Returns:
        TrainedSklearnClassifier or TrainedSklearnCVClassifier: The trained
            classifier object.
    """
    n_splits = cv if isinstance(cv, int) else cv.n_splits

    if save_path.exists():
        with save_path.open("rb") as file:
            return pickle.load(file)

    if is_trained(clf):
        return TrainedSklearnClassifier(clf, X=X)

    logger.info("Start training...")

    clf.fit(X, y)
    counts_per_class = np.unique(y, return_counts=True)[1]

    if min(counts_per_class) < n_splits:
        # Find the class with the minimum count
        min_class_index = np.argmin(counts_per_class)
        min_class_label = np.unique(y)[min_class_index]
        min_class_count = counts_per_class[min_class_index]
        logger.info(
            "Warning: Class '%s' has %d samples, less than 'n_splits' (=%d). "
            "Some cross-validation metrics won't be computed.",
            min_class_label,
            min_class_count,
            n_splits,
        )
    else:
        logger.info("Start cross-validation...")
        probabilities_cv = cross_val_predict(
            clf, X, y, cv=cv, method="predict_proba", n_jobs=n_jobs
        )
        metrics = cross_val_metrics(clf, X, y, probabilities_cv, cv)
        trained_clf = TrainedSklearnCVClassifier(
            clf=clf, probabilities_cv=probabilities_cv, X=X, metrics=metrics
        )

    with save_path.open("wb") as file:
        pickle.dump(trained_clf, file)

    return trained_clf


@dataclass
class ClassifierResult:
    """Data container for evaluation of classifier."""

    prediction: any
    model: any
    metrics: dict
    reports: dict


def fit_and_evaluate_clf(X, y, X_test, id_test, save_path, clf, cv, n_jobs=1):
    """Predicts the methylation class by supervised learning classifier.

    Uses supervised machine learning classifiers (Random Forest, K-Nearest
    Neighbors, Neural Networks, SVM, ...) to predict the methylation class of
    the sample. Output will be written to disk.

    Args:
        X (pd.DataFrame): Feature matrix (rows as samples, columns as
            features).
        y (array-like): Class labels.
        X_test (array-like): Value of the sample to be evaluated.
        id_test (str): Unique identifiers for the test samples to be evaluated.
        save_path (str or Path): Path where the classifiers and results
            will be saved/cached.
        clf (list): Classifier to use. Can be:

            - A scikit-learn classifier object or pipeline (trained or
              untrained).
            - A string in the format "scaler-selector-classifier". Possible
              values are:

            - A pipeline string composed of arbitrary components joined by
              dashes ("-"). Each component can be specified using either an
              abbreviation or the full class name (e.g., "std" or
              "StandardScaler").:

                scaler:
                    - "std": Standard scaling (StandardScaler).
                    - "minmax": Min-max scaling (MinMaxScaler).
                    - "robust": Robust scaling (RobustScaler).
                    - "power": Power transformation (PowerTransformer).
                    - "quantile": Quantile transformation
                      (QuantileTransformer).

                selector:
                    - "kbest": Select the best features (SelectKBest).
                    - "top": To varying features (TopVarianceSelector).
                    - "pca": Principal component analysis (PCA).
                    - "pca_auto": Principal component analysis (PCA). Number of
                      components is determined automatically.
                    - "lda": Linear Discriminant Analysis (LDA).

                clf:
                    - "rf": RandomForestClassifier.
                    - "lr": LogisticRegression.
                    - "et": ExtraTreesClassifier.
                    - "knn": KNeighborsClassifier.
                    - "mlp": MLPClassifier.
                    - "svc": Support Vector Classifier (SVC).
                    - "ada": AdaBoostClassifier.
                    - "bag": BaggingClassifier.
                    - "dt": DecisionTreeClassifier.
                    - "gp": GaussianProcessClassifier.
                    - "hgb": HistGradientBoostingClassifier.
                    - "nb": GaussianNB.
                    - "perceptron": Perceptron.
                    - "qda": Quadratic Discriminant Analysis (QDA).
                    - "ridge": RidgeClassifier.
                    - "sgd": SGDClassifier.


                Example: Using a feature selector and a classifier (SelectKBest
                    selection and Logistic Regression):
                    - "kbest-lr"

            - A custom class, that inherits from `TrainedClassifier`.

        cv (int or cross-validation generator): Determines the cross-validation
            splitting strategy.
        n_jobs (int): Number of parallel processes to run.

    Returns:
        ClassifierResult:
            - prediction (DataFrame): DataFrame containing the predicted
              probabilities for each class.
            - model (object): The trained classifier object.
            - metrics (dict): Dict containing classifier metrics.
            - reports (dict): Dict of evaluation report (both 'txt' and 'html')
              for each sample.
    """
    if isinstance(clf, (Pipeline, ClassifierMixin)):
        trained_clf = train_clf(
            clf=clf, X=X, y=y, save_path=save_path, cv=cv, n_jobs=n_jobs
        )
    elif isinstance(clf, str):
        pipeline = make_clf_pipeline(clf, X.shape, cv)
        trained_clf = train_clf(
            clf=pipeline, X=X, y=y, save_path=save_path, cv=cv, n_jobs=n_jobs
        )
    else:
        trained_clf = clf
    classes = trained_clf.classes()
    probabilities = trained_clf.predict_proba(X_test, id_test)
    info_txt = trained_clf.info(output_format="txt")
    info_html = trained_clf.info(output_format="html")
    prediction = pd.DataFrame(probabilities, index=id_test, columns=classes)
    reports_txt = make_reports(prediction, info_txt, output_format="txt")
    reports_html = make_reports(prediction, info_html, output_format="html")
    return ClassifierResult(
        prediction,
        trained_clf.model(),
        trained_clf.metrics(),
        {"txt": reports_txt, "html": reports_html},
    )



"""Module for handling Illumina array manifest files.

This module contains a single class ``Manifest`` for reading and processing
Illumina array manifest files, which contain information about probes and their
characteristics.
"""

import logging
import pickle
from pathlib import Path

import numpy as np
import pandas as pd
import pyranges as pr

from mepylome.dtypes.arrays import ArrayType
from mepylome.dtypes.cache import cache_key, input_args_id
from mepylome.dtypes.chromosome import Chromosome
from mepylome.dtypes.probes import Channel, InfiniumDesignType, ProbeType
from mepylome.utils.files import (
    download_file,
    ensure_directory_exists,
    get_csv_file,
    reset_file,
)
from mepylome.utils.varia import CONFIG, MEPYLOME_TMP_DIR

logger = logging.getLogger(__name__)

__all__ = ["Manifest"]


MANIFEST_DIR = Path.home() / ".mepylome" / "manifest_files_v0"
DOWNLOAD_DIR = MEPYLOME_TMP_DIR / "manifests"

ENDING_CONTROL_PROBES = CONFIG["suffixes"]["manifest_control_probes"]
NONE = -1

MANIFEST_URL = {
    ArrayType(type_): url for type_, url in CONFIG["urls"]["manifest"].items()
}
PROCESSED_MANIFEST_URL = (
    CONFIG["urls"]["processed_manifests"] % MANIFEST_DIR.name
)
REMOTE_FILENAME = {
    ArrayType(type_): url for type_, url in CONFIG["files"]["remote"].items()
}
LOCAL_FILENAME = {
    ArrayType(type_): url for type_, url in CONFIG["files"]["local"].items()
}


PROBES_COLUMNS = [
    "IlmnID",
    "Name",
    "AddressA_ID",
    "AddressB_ID",
    "Infinium_Design_Type",
    "Color_Channel",
    "CHR",
    "Chr",
    "MAPINFO",
    "AlleleA_ProbeSeq",
    "AlleleB_ProbeSeq",
]


CONTROL_COLUMNS = (
    "Address_ID",
    "Control_Type",
    "Color",
    "Extended_Type",
)


class Manifest:
    """Provides an object interface to an Illumina array manifest file.

    This class provides functionality for reading and processing Illumina array
    manifest files. A manifest can be loaded automatically based on the array
    type or provided as a raw manifest file. On first use, the necessary data
    is automatically downloaded if needed, transformed, and saved locally,
    which might take some time. The processed manifest is then saved locally
    and loaded in its processed form on subsequent uses. During a running
    session, all loaded manifests are cached in memory.

    Args:
        array_type (str or ArrayType): The type of array to process. Use either
            ArrayType (ArrayType.ILLUMINA_450K, ArrayType.ILLUMINA_EPIC,
            ArrayType.ILLUMINA_EPIC_V2) or corresponding string ('450k',
            'epic', 'epicv2', 'msa48')
        proc_path (str or Path): The path to the local processed manifest file
            (default: None).
        raw_path (str or Path, optional): Path to the raw manifest file.
            Default is None.
        download_proc (bool, optional): If True and there is no locally saved
            processed manifest file, attempts to download the processed
            manifest file instead of the raw one. Defaults to True.

    Examples:
        >>> # To initialize a manifest object for Illumina 450k array:
        >>> manifest = Manifest("450k")
        >>> manifest

        >>> # To initialize a manifest object for Illumina EPIC array
        >>> manifest = Manifest(ArrayType.ILLUMINA_EPIC)
        >>> type_1 = manifest.probe_info(ProbeType.ONE)

        >>> # To load all manifests when first used:
        >>> Manifest.load()
    """

    _cache = {}

    def __new__(
        cls,
        array_type=None,
        raw_path=None,
        proc_path=None,
        download_proc=True,
    ):
        key = cache_key(array_type, raw_path, proc_path)
        if key in cls._cache:
            return cls._cache[key]

        instance = super().__new__(cls)

        # Cache the instance
        cls._cache[key] = instance
        return instance

    def __getnewargs__(self):
        # Necessary for pickle
        return (
            self._init_array_type,
            self._init_raw_path,
            self._init_proc_path,
            self._init_download_proc,
        )

    def __init__(
        self,
        array_type=None,
        raw_path=None,
        proc_path=None,
        download_proc=True,
    ):
        if hasattr(self, "_cached"):
            return
        self._cached = True

        self._init_array_type = array_type
        self._init_raw_path = raw_path
        self._init_proc_path = proc_path
        self._init_download_proc = download_proc

        def to_path(x):
            return x if x is None else Path(x)

        self.array_type = ArrayType(array_type) if array_type else None
        self.raw_path = to_path(raw_path)
        self.proc_path = to_path(proc_path)
        self.download_proc = download_proc

        # Load cached data from disk
        pickle_hash = input_args_id(
            "manifest", self.array_type, self.raw_path, self.proc_path
        )
        self._pickle_path = MEPYLOME_TMP_DIR / f"{pickle_hash}.pkl"
        if self._pickle_path.exists():
            with self._pickle_path.open("rb") as file:
                saved_instance = pickle.load(file)
                self.__dict__.update(saved_instance.__dict__)
                return

        if self.array_type == ArrayType.UNKNOWN:
            self._data_frame = pd.DataFrame()
            self._control_data_frame = pd.DataFrame()
            self._snp_data_frame = pd.DataFrame()
            self._methyl_probes = None
            return

        # Set processed manifest path
        if self.proc_path is None:
            if self.array_type is not None:
                self.proc_path = MANIFEST_DIR / LOCAL_FILENAME[self.array_type]
            elif self.raw_path is not None:
                if self.raw_path.suffix == ".zip":
                    gz_filename = "proc_" + self.raw_path.stem + ".gz"
                else:
                    gz_filename = "proc_" + self.raw_path.name + ".gz"
                self.proc_path = DOWNLOAD_DIR / gz_filename
            else:
                msg = "Provide either array_type or proc_path or raw_path"
                raise ValueError(msg)

        self.ctrl_path = Manifest._get_control_path(self.proc_path)

        # Create processed manifest files if they do not exist
        if not (self.proc_path.exists() and self.ctrl_path.exists()):
            # If array type is given, download the files
            if self.array_type is not None:
                if (
                    not self.download_proc
                    or not self._download_processed_manifest()
                ):
                    self._download_manifest()
                    csv_filename = REMOTE_FILENAME[self.array_type]
                    self._process_manifest(csv_filename=csv_filename)
            # Else process from raw_path
            elif self.raw_path is not None:
                if self.raw_path.suffix == ".zip":
                    csv_filename = self.raw_path.stem
                else:
                    csv_filename = self.raw_path.name
                self._process_manifest(csv_filename=csv_filename)
            else:
                msg = "Provide either array_type or proc_path or raw_path"
                raise ValueError(msg)

        self._data_frame = self._read_probes(self.proc_path)
        self._control_data_frame = self._read_control_probes(self.ctrl_path)
        self._snp_data_frame = self._read_snp_probes()
        self._methyl_probes = None

        # Save to disk
        with self._pickle_path.open("wb") as file:
            pickle.dump(self, file)

    @property
    def data_frame(self):
        """Pandas data frame of all manifest probes."""
        return self._data_frame

    @property
    def control_data_frame(self):
        """Pandas data frame of all manifest control probes."""
        return self._control_data_frame

    @property
    def snp_data_frame(self):
        """SNP probes from the manifest data frame."""
        return self._snp_data_frame

    @property
    def methylation_probes(self):
        """All type I and II probes."""
        if self._methyl_probes is None:
            type_1 = self.probe_info(ProbeType.ONE)
            type_2 = self.probe_info(ProbeType.TWO)
            idx = np.sort(
                np.concatenate(
                    [
                        type_1.IlmnID.index,
                        type_2.IlmnID.index,
                    ]
                )
            )
            self._methyl_probes = self._data_frame.iloc[idx]["IlmnID"].values
        return self._methyl_probes

    def control_address(self, control_type=None):
        """Returns address IDs of all control probes of the specified type."""
        if control_type is None:
            return self._control_data_frame.Address_ID
        # Ensure control_type is a list-like object
        if not isinstance(control_type, (list, tuple)):
            control_type = [control_type]
        # Use isin() with the list-like object
        return self._control_data_frame[
            self._control_data_frame.Control_Type.isin(control_type)
        ].Address_ID

    @staticmethod
    def load(array_types=None):
        """Loads specified manifests into memory.

        Args:
            array_types (list or ArrayType, optional): List of array types or
                a single array type to load. Defaults to all available types.

        Examples:
            >>> # Load all manifests:
            >>> Manifest.load()

            >>> # Load specific manifests:
            >>> Manifest.load(
            >>>     [ArrayType.ILLUMINA_450K, ArrayType.ILLUMINA_EPIC]
            >>> )
            >>> Manifest.load("epicv2")
        """
        if array_types is None:
            array_types = [
                ArrayType.ILLUMINA_450K,
                ArrayType.ILLUMINA_EPIC,
                ArrayType.ILLUMINA_EPIC_V2,
                ArrayType.ILLUMINA_MSA48,
            ]
        if not isinstance(array_types, list):
            array_types = [array_types]
        for array_type in array_types:
            _ = Manifest(array_type)

    def _download_manifest(
        self,
    ):
        """Downloads the appropriate manifest file.

        This method downloads the manifest file based on the `array_type`
        specified during the initialization of the Manifest object. It stores
        the downloaded file in the path of the `raw_path` attribute.

        Raises:
            KeyError: If the array type is not found in MANIFEST_URL.
        """
        source_url = MANIFEST_URL[self.array_type]
        source_filename = Path(source_url).name

        logger.info("Downloading manifest: %s", source_filename)

        self.raw_path = DOWNLOAD_DIR / source_filename
        download_file(source_url, self.raw_path)

    def _download_processed_manifest(self):
        """Download processed manifest file and return true if successful."""
        logger.info("Downloading processed %s manifest", self.array_type)
        try:
            for path in [self.proc_path, self.ctrl_path]:
                ensure_directory_exists(path.parent)
                url = PROCESSED_MANIFEST_URL + path.name
                download_file(url, path)
            return True
        except Exception:
            return False

    @staticmethod
    def _get_control_path(probes_path):
        """Converts probes path to the control probes path."""
        split_filename = probes_path.name.split(".")
        split_filename[0] += ENDING_CONTROL_PROBES
        return Path(probes_path.parent, ".".join(split_filename))

    def _process_manifest(self, csv_filename=None):
        """Process the manifest file and save it locally to disk.

        This method processes the raw manifest file by extracting the necessary
        probe and control probe information, and then saves these processed
        details to local files with pathnames `probes_path` and `ctrl_path`.

        Args:
            csv_filename (str, optional): Name of the manifest file inside the
                archive. If not provided, it defaults to the name of the
                raw_path file.
        """
        logger.info("Process raw manifest %s", self.raw_path)
        if csv_filename is None:
            csv_filename = self.raw_path.name
        ensure_directory_exists(self.proc_path.parent)
        ensure_directory_exists(self.ctrl_path.parent)
        with get_csv_file(self.raw_path, csv_filename) as manifest_file:
            # Process probes
            Manifest._seek_to_start(manifest_file)
            available_columns = pd.read_csv(manifest_file, nrows=0).columns
            Manifest._seek_to_start(manifest_file)
            valid_columns = [
                col for col in PROBES_COLUMNS if col in available_columns
            ]
            probes_df = pd.read_csv(
                manifest_file,
                low_memory=False,
                usecols=valid_columns,
            )
            n_probes = probes_df[probes_df.IlmnID.str.startswith("[")].index[0]
            probes_df = probes_df[:n_probes]
            probes_df = Manifest._process_probes(probes_df)
            probes_df.to_csv(self.proc_path, index=False)
            # Process controls
            Manifest._seek_to_start(manifest_file)
            controls_df = pd.read_csv(
                manifest_file,
                header=None,
                # Skip metadata and probe section
                skiprows=(3 + n_probes),
                usecols=range(len(CONTROL_COLUMNS)),
            )
            controls_df.columns = CONTROL_COLUMNS
            if (
                pd.to_numeric(controls_df["Address_ID"], errors="coerce")
                .notna()
                .all()
            ):
                controls_df["Address_ID"] = controls_df["Address_ID"].astype(
                    "int32"
                )
            controls_df.to_csv(self.ctrl_path, index=False)

    @staticmethod
    def _process_probes(data_frame):
        """Transforms manifest probes to a more efficient internal format."""
        rename_map = {
            "MAPINFO": "Start",
            "CHR": "Chromosome",
            "Chr": "Chromosome",
        }
        data_frame = data_frame.rename(columns=rename_map)
        data_frame["Chromosome"] = Chromosome.pd_from_string(
            data_frame["Chromosome"]
        )
        # IlmnID and Name are different in EPICv2
        data_frame["IlmnID"] = data_frame["Name"]
        data_frame = data_frame.drop(columns=["Name"])
        channel_to_int = {"Grn": Channel.GRN, "Red": Channel.RED}
        design_type_map = {
            "I": InfiniumDesignType.I,
            "II": InfiniumDesignType.II,
        }
        with pd.option_context("future.no_silent_downcasting", True):
            if "Color_Channel" in data_frame.columns:
                data_frame["Color_Channel"] = (
                    data_frame["Color_Channel"]
                    .replace(channel_to_int)
                    .infer_objects()
                )

            if "Infinium_Design_Type" in data_frame.columns:
                data_frame["Infinium_Design_Type"] = (
                    data_frame["Infinium_Design_Type"]
                    .replace(design_type_map)
                    .infer_objects()
                )

        data_frame["TypeI_N_CpG"] = 0
        data_frame["TypeII_N_CpG"] = 0

        if "AlleleB_ProbeSeq" in data_frame.columns:
            data_frame["TypeI_N_CpG"] = np.maximum(
                0,
                data_frame["AlleleB_ProbeSeq"].fillna("").str.count("CG") - 1,
            )

        if "AlleleA_ProbeSeq" in data_frame.columns:
            # R Stands for CG in AlleleA_ProbeSeq
            data_frame["TypeII_N_CpG"] = (
                data_frame["AlleleA_ProbeSeq"].fillna("").str.count("R")
            )

        data_frame["N_CpG"] = NONE
        if "Infinium_Design_Type" in data_frame.columns:
            is_type_I = (
                data_frame["Infinium_Design_Type"] == InfiniumDesignType.I
            )
            is_type_II = (
                data_frame["Infinium_Design_Type"] == InfiniumDesignType.II
            )

            data_frame.loc[is_type_I, "N_CpG"] = data_frame.loc[
                is_type_I, "TypeI_N_CpG"
            ]
            data_frame.loc[is_type_II, "N_CpG"] = data_frame.loc[
                is_type_II, "TypeII_N_CpG"
            ]

        # Use int32 to improve performance of indexing
        int_cols = [
            "AddressA_ID",
            "AddressB_ID",
            "Infinium_Design_Type",
            "Start",
            "Color_Channel",
            "TypeI_N_CpG",
            "TypeII_N_CpG",
        ]
        for col in int_cols:
            if col in data_frame.columns:
                data_frame[col] = data_frame[col].fillna(NONE).astype("int32")

        if "Start" in data_frame.columns:
            data_frame["End"] = data_frame["Start"]

        drop_cols = ["AlleleA_ProbeSeq", "AlleleB_ProbeSeq"]
        existing_drop_cols = [
            col for col in drop_cols if col in data_frame.columns
        ]
        data_frame = data_frame.drop(columns=existing_drop_cols)

        def get_probe_type(name, infinium_type):
            """Determines the probe type (I, II, SnpI, SnpII or Control)."""
            probe_type = ProbeType.from_manifest_values(name, infinium_type)
            return probe_type.value

        if (
            "IlmnID" in data_frame.columns
            and "Infinium_Design_Type" in data_frame.columns
        ):
            vectorized_get_type = np.vectorize(get_probe_type)
            data_frame["Probe_Type"] = vectorized_get_type(
                data_frame["IlmnID"].values,
                data_frame["Infinium_Design_Type"].values,
            )

        # TODO: drop Infinium_Design_Type
        if not {"Chromosome", "Start", "End"}.issubset(data_frame.columns):
            return data_frame

        probes_ranges = pr.PyRanges(data_frame).sort()
        return probes_ranges.df

    @staticmethod
    def _seek_to_start(manifest_file):
        """Move the manifest file pointer to the start of the data section.

        Details:
            The function searches for the first occurrence of the left-most
            column named "IlmnID".
        """
        reset_file(manifest_file)

        current_pos = manifest_file.tell()
        header_line = manifest_file.readline()

        while not header_line.startswith(b"IlmnID"):
            current_pos = manifest_file.tell()
            if not header_line:
                msg = (
                    "The first (left-most) column in your manifest must "
                    "contain 'IlmnID'. This defines the header row."
                )
                raise EOFError(msg)
            header_line = manifest_file.readline()

        if current_pos == 0:
            manifest_file.seek(current_pos)
        else:
            manifest_file.seek(current_pos - 1)

    def _read_probes(self, probes_file):
        """Reads and returns probes from local file `probes_file`."""
        data_frame = pd.read_csv(
            probes_file,
            dtype=self._get_data_types(),
        )
        # TODO: epicv2 has duplicated ID's for example c='cg22367159'
        data_frame = data_frame.drop_duplicates(
            subset=["IlmnID"], keep="first"
        )
        return data_frame.reset_index(drop=True)

    def _read_control_probes(self, control_file):
        """Reads and returns control probes from local file `control_file`."""
        data_frame = pd.read_csv(
            control_file,
            dtype=self._get_data_types(),
        )
        # Use int32 instead of int64 to improve performance of indexing
        if (
            pd.to_numeric(data_frame["Address_ID"], errors="coerce")
            .notna()
            .all()
        ):
            data_frame["Address_ID"] = (
                data_frame["Address_ID"].astype("int32").copy()
            )
        return data_frame

    def _read_snp_probes(self):
        """Extracts SNP probes from the manifest data frame."""
        snp_df = self.data_frame.copy()
        return snp_df[snp_df.IlmnID.str.match("rs", na=False)]

    def _get_data_types(self):
        """Returns data types for the manifest columns."""
        return {
            "IlmnID": np.dtype(object),
            "AddressA_ID": np.int32,
            "AddressB_ID": np.int32,
            "Infinium_Design_Type": np.int8,
            "Color_Channel": np.int8,
            "Chromosome": np.int8,
            "Start": np.int32,
            "TypeI_N_CpG": np.int8,
            "TypeII_N_CpG": np.int8,
            "N_CpG": np.int8,
            "End": np.int32,
            "Probe_Type": np.int8,
        }

    def probe_info(self, probe_type, channel=None):
        """Retrieves information about probes of a specified type and channel.

        Args:
            probe_type (ProbeType): The type of probe (I, II, SnpI, SnpII,
                Control).
            channel (Channel, optional): The color channel (RED or GRN).
                Defaults to None.

        Returns:
            DataFrame: DataFrame containing information about the specified
                probes.

        Raises:
            ValueError: If probe_type is not a valid ProbeType or if channel is
            not a valid Channel.
        """
        if not isinstance(probe_type, ProbeType):
            msg = "probe_type is not a valid ProbeType"
            raise TypeError(msg)

        if channel and not isinstance(channel, Channel):
            msg = "channel not a valid Channel"
            raise TypeError(msg)

        data_frame = self.data_frame
        probe_type_mask = data_frame["Probe_Type"].values == probe_type.value

        if channel is None:
            return data_frame[probe_type_mask]

        channel_mask = data_frame["Color_Channel"].values == channel.value
        return data_frame[probe_type_mask & channel_mask]

    def __repr__(self):
        title = f"Manifest({self.array_type}):"
        lines = [
            title + "\n" + "*" * len(title),
            f"data_frame:\n{self.data_frame}",
            f"control_data_frame:\n{self.control_data_frame}",
            f"snp_data_frame:\n{self.snp_data_frame}",
        ]
        return "\n\n".join(lines)



"""Contains the ArrayType used to represent different Illumina array types."""

from enum import Enum, unique
from pathlib import Path

from mepylome.dtypes.idat import IdatParser


def _find_valid_path(path):
    """Tries to find a valid IDAT file associated with the given path."""
    base_path = Path(path)
    if base_path.exists():
        return base_path
    grn_path = Path(str(base_path) + "_Grn.idat")
    if grn_path.exists():
        return grn_path
    grn_gz_path = Path(str(base_path) + "_Grn.idat.gz")
    if grn_gz_path.exists():
        return grn_gz_path
    msg = f"No valid file found for path: {path}"
    raise ValueError(msg)


@unique
class ArrayType(Enum):
    """Provides constants for the different Illumina array types.

    Enum representing different Illumina array types with method to infer
    type from probe count.
    """

    ILLUMINA_27K = "27k"
    ILLUMINA_450K = "450k"
    ILLUMINA_EPIC = "epic"
    ILLUMINA_EPIC_V2 = "epicv2"
    ILLUMINA_MSA48 = "msa48"
    ILLUMINA_MOUSE = "mouse"
    UNKNOWN = "unknown"

    def __str__(self):
        return self.value

    @classmethod
    def from_probe_count(cls, probe_count):
        """Infers array type based on the number of probes in an idat file."""
        if 622000 <= probe_count <= 623000:
            return cls.ILLUMINA_450K

        if 1050000 <= probe_count <= 1053000:
            return cls.ILLUMINA_EPIC

        if 1032000 <= probe_count <= 1033000:
            return cls.ILLUMINA_EPIC

        if 1100000 <= probe_count <= 1108000:
            return cls.ILLUMINA_EPIC_V2

        if 384400 <= probe_count <= 384600:
            return cls.ILLUMINA_MSA48

        if 55200 <= probe_count <= 55400:
            return cls.ILLUMINA_27K

        if 315000 <= probe_count <= 362000:
            return cls.ILLUMINA_MOUSE

        return cls.UNKNOWN

    @classmethod
    def from_idat(cls, path):
        """Infers array type from idat_file."""
        valid_path = _find_valid_path(path)
        probe_count = IdatParser(valid_path, array_type_only=True).n_snps_read
        return ArrayType.from_probe_count(probe_count)
